# -*- coding: utf-8 -*-
"""
This module contains functions for common path operations.
"""

# system imports
import os
import os.path as osp
import errno
import shutil
import itertools
import unicodedata
from stat import S_ISDIR
from typing import List, Optional, Tuple, Callable, Iterator, Iterable, Union

# local imports
from .content_hasher import DropboxContentHasher


def _path_components(path: str) -> List[str]:
    components = path.strip(osp.sep).split(osp.sep)
    cleaned_components = [c for c in components if c]
    return cleaned_components


def normalize_case(string: str) -> str:
    """
    Converts a string to lower case. Todo: Follow Python 2.5 / Dropbox conventions.

    :param string: Original string.
    :returns: Lowercase string.
    """
    return string.lower()


def normalize_unicode(string: str) -> str:
    """
    Normalizes a string to replace all decomposed unicode characters with their single
    character equivalents.

    :param string: Original string.
    :returns: Normalized string.
    """
    return unicodedata.normalize("NFC", string)


def normalize(string: str) -> str:
    """
    Replicates the path normalization performed by Dropbox servers. This typically only
    involves converting the path to lower case, with a few (undocumented) exceptions:

    * Unicode normalization: decomposed characters are converted to composed characters.
    * Lower casing of non-ascii characters: Dropbox uses the Python 2.5 behavior for
      conversion to lower case. This means that some cyrillic characters are incorrectly
      lower-cased. For example:
      "Ꙋ".lower() -> "Ꙋ" instead of "ꙋ"
      "ΣΣΣ".lower() -> "σσσ" instead of "σσς"
    * Trailing spaces are stripped from folder names. We do not perform this
      normalization here because the Dropbox API will raise sync errors for such names
      anyways.

    Note that calling :func:`normalize` on an already normalized path will return the
    unmodified input.

    Todo: Follow Python 2.5 / Dropbox conventions instead of Python 3 conventions.

    :param string: Original path.
    :returns: Normalized path.
    """
    return normalize_case(normalize_unicode(string))


def is_fs_case_sensitive(path: str) -> bool:
    """
    Checks if ``path`` lies on a partition with a case-sensitive file system.

    :param path: Path to check.
    :returns: Whether ``path`` lies on a partition with a case-sensitive file system.
    """
    if path.islower():
        check_path = path.upper()
    else:
        check_path = path.lower()

    if osp.exists(path) and not osp.exists(check_path):
        return True
    else:
        return not osp.samefile(path, check_path)


def is_child(path: str, parent: str) -> bool:
    """
    Checks if ``path`` semantically is inside ``parent``. Neither path needs to
    refer to an actual item on the drive. This function is case sensitive.

    :param path: Item path.
    :param parent: Parent path.
    :returns: Whether ``path`` semantically lies inside ``parent``.
    """

    parent = parent.rstrip(osp.sep) + osp.sep
    path = path.rstrip(osp.sep)

    return path.startswith(parent)


def is_equal_or_child(path: str, parent: str) -> bool:
    """
    Checks if ``path`` semantically is inside ``parent`` or equals ``parent``. Neither
    path needs to refer to an actual item on the drive. This function is case sensitive.

    :param path: Item path.
    :param parent: Parent path.
    :returns: ``True`` if ``path`` semantically lies inside ``parent`` or
        ``path == parent``.
    """

    return is_child(path, parent) or path == parent


def equivalent_path_candidates(
    path: str,
    root: str = osp.sep,
    norm_func: Callable = normalize,
) -> List[str]:
    """
    Given a "normalized" path using an injective (one-directional) normalization
    function, this method returns a list of matching un-normalized local paths.

    If no such local path exists, the normalized path itself is returned. If a local
    path can be followed up to a certain parent in the hierarchy, it will be taked and
    the remaining normalized path will be appended.

    :Example:

        Assume the normalization function is ``str.lower()``. If a root directory
        contains two folders "/parent/subfolder/child" and "/parent/Subfolder/child",
        two matches will be returned for "path = /parent/subfolder/child/file.txt".

    :param path: Normalized path relative to ``root``.
    :param root: Parent directory to search in. There are significant performance
        improvements if a root directory with a small tree is given.
    :param norm_func: Normalization function to use. Defaults to :func:`normalize`.
    :returns: Candidates for correctly cased local paths.
    """

    path = path.lstrip(osp.sep)

    if path == "":
        return [root]

    components = _path_components(path)
    n_components_root = len(_path_components(root))

    candidates = {-1: [root]}

    for root, dirs, files in os.walk(root):

        n_components_current_root = len(_path_components(root))
        depth = n_components_current_root - n_components_root

        all_dirs = dirs.copy()
        all_files = files.copy()

        dirs.clear()
        files.clear()

        if depth >= len(components):
            # Current path is too deep to be match, skip it.
            continue

        dirname_normalized = norm_func(components[depth])

        for dirname in all_dirs:
            if norm_func(dirname) == dirname_normalized:
                dirs.append(dirname)

        if depth + 1 == len(components):
            # Any matching entries must be direct children of root: check files.
            for filename in all_files:
                if norm_func(filename) == dirname_normalized:
                    files.append(filename)

        new_candidates = [osp.join(root, name) for name in itertools.chain(dirs, files)]

        if new_candidates:
            candidates[depth] = candidates.get(depth, []) + new_candidates

    i_max = max(candidates.keys())
    best_candidates = candidates[i_max]
    local_paths = [osp.join(path, *components[i_max + 1 :]) for path in best_candidates]

    return local_paths


def denormalize_path(path: str, root: str = osp.sep) -> str:
    """
    Returns a denormalized version of the given path as far as corresponding nodes with
    the same normalization exist in the given root directory. If multiple matches are
    found, only one is returned. If ``path`` does not exist in root ``root`` or ``root``
    does not exist, the return value will be ``os.path.join(root, path)``.

    :param path: Original path relative to ``root``.
    :param root: Parent directory to search in. There are significant performance
        improvements if a root directory with a small tree is given.
    :returns: Absolute and cased version of given path.
    """

    candidates = equivalent_path_candidates(path, root)
    return candidates[0]


def to_existing_unnormalized_path(path: str, root: str = osp.sep) -> str:
    """
    Returns a cased version of the given path if corresponding nodes (with arbitrary
    casing) exist in the given root directory. If multiple matches are found, only one
    is returned.

    :param path: Original path relative to ``root``.
    :param root: Parent directory to search in. There are significant performance
        improvements if a root directory with a small tree is given.
    :returns: Absolute and cased version of given path.
    :raises FileNotFoundError: if ``path`` does not exist in root ``root`` or ``root``
        itself does not exist.
    """

    candidates = equivalent_path_candidates(path, root)

    for candidate in candidates:
        if osp.exists(candidate):
            return candidate

    raise FileNotFoundError(f'No matches with different casing found in "{root}"')


def normalized_path_exists(path: str, root: str = osp.sep) -> bool:
    """
    Checks if a ``path`` exists in given ``root`` directory, similar to
    ``os.path.exists`` but case-insensitive. Normalisation is performed as by Dropbox
    servers (lower case and unicode normalisation).

    :param path: Path relative to ``root``.
    :param root: Directory where we will look for ``path``. There are significant
        performance improvements if a root directory with a small tree is given.
    :returns: Whether an arbitrarily cased version of ``path`` exists.
    """

    candidates = equivalent_path_candidates(path, root)

    for c in candidates:
        if osp.exists(c):
            return True

    return False


def generate_cc_name(path: str, suffix: str = "conflicting copy") -> str:
    """
    Generates a path for a conflicting copy of ``path``. The file name is created by
    inserting the given ``suffix`` between the the filename and extension. For instance:

        "my_file.txt" -> "my_file (conflicting copy).txt"

    If a file with the resulting path already exists (case-insensitive!), we
    additionally append an integer number, for instance:

        "my_file.txt" -> "my_file (conflicting copy 1).txt"

    :param path: Original path name.
    :param suffix: Suffix to use. Defaults to "conflicting copy".
    :returns: New path.
    """

    dirname, basename = osp.split(path)
    filename, ext = osp.splitext(basename)

    i = 0
    cc_candidate = f"{filename} ({suffix}){ext}"

    while normalized_path_exists(cc_candidate, dirname):
        i += 1
        cc_candidate = f"{filename} ({suffix} {i}){ext}"

    return osp.join(dirname, cc_candidate)


def delete(path: str, raise_error: bool = False) -> Optional[OSError]:
    """
    Deletes a file or folder at ``path``.

    :param path: Path of item to delete.
    :param raise_error: Whether to raise errors or return them.
    :returns: Any caught exception during the deletion.
    """
    err = None

    try:
        shutil.rmtree(path)
    except OSError:
        try:
            os.unlink(path)
        except OSError as e:
            err = e

    if raise_error and err:
        raise err
    else:
        return err


def move(
    src_path: str,
    dest_path: str,
    raise_error: bool = False,
    preserve_dest_permissions=False,
) -> Optional[OSError]:
    """
    Moves a file or folder from ``src_path`` to ``dest_path``. If either the source or
    the destination path no longer exist, this function does nothing. Any other
    exceptions are either raised or returned if ``raise_error`` is False.

    :param src_path: Path of item to move.
    :param dest_path: Destination path. Any existing file at this path will be replaced
        by the move. Any existing **empty** folder will be replaced if the source is
        also a folder.
    :param raise_error: Whether to raise errors or return them.
    :param preserve_dest_permissions: Whether to apply the permissions of the source
        path to the destination path. Permissions will not be set recursively.
    :returns: Any caught exception during the move.
    """

    err: Optional[OSError] = None
    orig_mode: Optional[int] = None

    if preserve_dest_permissions:
        # save dest permissions
        try:
            orig_mode = os.stat(dest_path).st_mode & 0o777
        except FileNotFoundError:
            pass

    try:
        shutil.move(src_path, dest_path)
    except FileNotFoundError:
        # do nothing of source or dest path no longer exist
        pass
    except OSError as exc:
        err = exc
    else:
        if orig_mode:
            # reapply dest permissions
            try:
                os.chmod(dest_path, orig_mode)
            except OSError:
                pass

    if raise_error and err:
        raise err
    else:
        return err


def walk(
    root: Union[str, os.PathLike],
    listdir: Callable[[Union[str, os.PathLike]], Iterable[os.DirEntry]] = os.scandir,
) -> Iterator[Tuple[str, os.stat_result]]:
    """
    Iterates recursively over the content of a folder.

    :param root: Root folder to walk.
    :param listdir: Function to call to get the folder content.
    :returns: Iterator over (path, stat) results.
    """

    for entry in listdir(root):

        try:
            path = entry.path
            stat = entry.stat()

            yield path, stat

            if S_ISDIR(stat.st_mode):
                for res in walk(entry.path, listdir=listdir):
                    yield res

        except OSError as exc:
            # Directory may have been deleted between finding it in the directory
            # list of its parent and trying to list its contents. If this
            # happens we treat it as empty. Likewise if the directory was replaced
            # with a file of the same name (less likely, but possible).
            if exc.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
                return
            else:
                raise


def content_hash(
    local_path: str, chunk_size: int = 1024
) -> Tuple[Optional[str], Optional[float]]:
    """
    Computes content hash of a local file.

    :param local_path: Absolute path on local drive.
    :param chunk_size: Size of chunks to hash in bites.
    :returns: Content hash to compare with Dropbox's content hash and mtime just before
        the hash was computed.
    """

    hasher = DropboxContentHasher()

    try:
        mtime = os.stat(local_path).st_mtime

        try:
            with open(local_path, "rb") as f:
                while True:
                    chunk = f.read(chunk_size)
                    if len(chunk) == 0:
                        break
                    hasher.update(chunk)

        except IsADirectoryError:
            return "folder", mtime
        else:
            return str(hasher.hexdigest()), mtime

    except FileNotFoundError:
        return None, None
    except NotADirectoryError:
        # a parent directory in the path refers to a file instead of a folder
        return None, None
    finally:
        del hasher



# -*- coding: utf-8 -*-
"""
This module defines functions to start and stop the sync daemon and retrieve proxy
objects for a running daemon.
"""

# system imports
import sys
import os
import time
import signal
import enum
import subprocess
import threading
import fcntl
import struct
import warnings
import argparse
import ast
import re
from pprint import pformat
from shlex import quote
from typing import Optional, Any, Union, Tuple, Dict, Iterable, Type, TYPE_CHECKING
from types import TracebackType

# external imports
import Pyro5  # type: ignore
from Pyro5.errors import CommunicationError  # type: ignore
from Pyro5.api import Daemon, Proxy, expose, oneway, register_dict_to_class  # type: ignore
import sdnotify  # type: ignore
from fasteners import InterProcessLock  # type: ignore

# local imports
from .errors import SYNC_ERRORS, GENERAL_ERRORS, MaestralApiError
from .utils import exc_info_tuple
from .utils.appdirs import get_runtime_path
from .constants import IS_MACOS, ENV


if TYPE_CHECKING:
    from .main import Maestral


__all__ = [
    "Stop",
    "Start",
    "Lock",
    "maestral_lock",
    "get_maestral_pid",
    "sockpath_for_config",
    "lockpath_for_config",
    "wait_for_startup",
    "is_running",
    "freeze_support",
    "start_maestral_daemon",
    "start_maestral_daemon_process",
    "stop_maestral_daemon_process",
    "MaestralProxy",
    "CommunicationError",
]


# systemd environment
INVOCATION_ID = os.getenv("INVOCATION_ID")
NOTIFY_SOCKET = os.getenv("NOTIFY_SOCKET")
WATCHDOG_PID = os.getenv("WATCHDOG_PID")
WATCHDOG_USEC = os.getenv("WATCHDOG_USEC")
IS_WATCHDOG = WATCHDOG_USEC and (
    WATCHDOG_PID is None or int(WATCHDOG_PID) == os.getpid()
)


URI = "PYRO:maestral.{0}@{1}"
Pyro5.config.THREADPOOL_SIZE_MIN = 2


def freeze_support() -> None:
    """
    Call this as early as possible in the main entry point of a frozen executable.
    This call will start the sync daemon if a matching command line arguments are
    detected and do nothing otherwise.
    """

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("-c")
    parsed_args, _ = parser.parse_known_args()

    if parsed_args.c:
        template = r'.*start_maestral_daemon\("(?P<config_name>\S+)", start_sync=(?P<start_sync>\S+)\).*'
        match = re.match(template, parsed_args.c)

        if match:
            config_name = match["config_name"]
            start_sync = ast.literal_eval(match["start_sync"])

            start_maestral_daemon(config_name, start_sync)
            sys.exit()


class Stop(enum.Enum):
    """Enumeration of daemon exit results"""

    Ok = 0
    Killed = 1
    NotRunning = 2
    Failed = 3


class Start(enum.Enum):
    """Enumeration of daemon start results"""

    Ok = 0
    AlreadyRunning = 1
    Failed = 2


# ==== error serialization =============================================================


def serpent_deserialize_api_error(class_name: str, d: dict) -> MaestralApiError:
    """
    Deserializes a MaestralApiError.

    :param class_name: Name of class to deserialize.
    :param d: Dictionary of serialized class.
    :returns: Class instance.
    """

    # Import maestral.errors for evaluation.
    # This import needs to be absolute to reconstruct the Exception class. Note that the
    # eval is safe here because ``serpent_deserialize_api_error`` is only registered for
    # strings that match an error class name.

    import maestral.errors  # noqa: F401

    cls = eval(class_name)
    err = cls(*d["args"])
    for a_name, a_value in d["attributes"].items():
        setattr(err, a_name, a_value)

    return err


for err_cls in (*SYNC_ERRORS, *GENERAL_ERRORS):
    register_dict_to_class(
        err_cls.__module__ + "." + err_cls.__name__, serpent_deserialize_api_error
    )


# ==== interprocess locking ============================================================


def _get_lockdata() -> Tuple[bytes, str, int]:

    try:
        os.O_LARGEFILE
    except AttributeError:
        start_len = "ll"
    else:
        start_len = "qq"

    if sys.platform.startswith(("netbsd", "freebsd", "openbsd")) or IS_MACOS:
        if struct.calcsize("l") == 8:
            off_t = "l"
            pid_t = "i"
        else:
            off_t = "lxxxx"
            pid_t = "l"

        fmt = off_t + off_t + pid_t + "hh"
        pid_index = 2
        lockdata = struct.pack(fmt, 0, 0, 0, fcntl.F_WRLCK, 0)
    # elif sys.platform.startswith('gnukfreebsd'):
    #     fmt = 'qqihhi'
    #     pid_index = 2
    #     lockdata = struct.pack(fmt, 0, 0, 0, fcntl.F_WRLCK, 0, 0)
    # elif sys.platform in ('hp-uxB', 'unixware7'):
    #     fmt = 'hhlllii'
    #     pid_index = 2
    #     lockdata = struct.pack(fmt, fcntl.F_WRLCK, 0, 0, 0, 0, 0, 0)
    elif sys.platform.startswith("linux"):
        fmt = "hh" + start_len + "ih"
        pid_index = 4
        lockdata = struct.pack(fmt, fcntl.F_WRLCK, 0, 0, 0, 0, 0)
    else:
        raise RuntimeError(f"Unsupported platform {sys.platform}")

    return lockdata, fmt, pid_index


class Lock:
    """A inter-process and inter-thread lock

    This internally uses :class:`fasteners.InterProcessLock` but provides non-blocking
    acquire. It also guarantees thread-safety when using the :meth:`singleton` class
    method to create / retrieve a lock instance.

    :param path: Path of the lock file to use / create.
    """

    _instances: Dict[str, "Lock"] = {}
    _singleton_lock = threading.Lock()

    @classmethod
    def singleton(cls, path: str) -> "Lock":
        """
        Retrieve an existing lock object with a given 'name' or create a new one. Use
        this method for thread-safe locks.

        :param path: Path of the lock file to use / create.
        """

        with cls._singleton_lock:
            try:
                instance = cls._instances[path]
            except KeyError:
                instance = cls(path)
                cls._instances[path] = instance

            return instance

    def __init__(self, path: str) -> None:

        self.path = path

        self._internal_lock = threading.Semaphore()
        self._external_lock = InterProcessLock(self.path)

        self._lock = threading.RLock()
        self._acquired = False  # tracks if we acquired the lock

    def acquire(self) -> bool:
        """
        Attempts to acquire the given lock.

        :returns: Whether or not the acquisition succeeded.
        """

        with self._lock:
            locked_internal = self._internal_lock.acquire(blocking=False)

            if not locked_internal:
                return False

            try:
                locked_external = self._external_lock.acquire(blocking=False)
            except Exception:
                self._internal_lock.release()
                raise
            else:

                if locked_external:
                    self._acquired = True
                    return True
                else:
                    self._internal_lock.release()
                    return False

    def release(self) -> None:
        """Release the previously acquired lock."""
        with self._lock:

            if not self._acquired:
                raise RuntimeError(
                    "Cannot release a lock, it was acquired by a different process"
                )

            self._external_lock.release()
            self._internal_lock.release()
            self._acquired = False

    def locked(self) -> bool:
        """
        Checks if the lock is currently held by any thread or process.

        :returns: Whether the lock is acquired.
        """
        with self._lock:
            gotten = self.acquire()
            if gotten:
                self.release()
            return not gotten

    def locking_pid(self) -> Optional[int]:
        """
        Returns the PID of the process which currently holds the lock or ``None``. This
        should work on macOS, OpenBSD and Linux but may fail on some platforms. Always
        use :meth:`locked` to check if the lock is held by any process.

        :returns: The PID of the process which currently holds the lock or ``None``.
        """

        with self._lock:

            if self._acquired:
                return os.getpid()

            try:
                # Don't close again in case we are the locking process.
                self._external_lock._do_open()
                lockdata, fmt, pid_index = _get_lockdata()
                lockdata = fcntl.fcntl(
                    self._external_lock.lockfile, fcntl.F_GETLK, lockdata
                )

                lockdata_list = struct.unpack(fmt, lockdata)
                pid = lockdata_list[pid_index]

                if pid > 0:
                    return pid

            except OSError:
                pass

            return None


# ==== helpers for daemon management ===================================================


def _send_term(pid: int) -> None:
    try:
        os.kill(pid, signal.SIGTERM)
    except ProcessLookupError:
        pass


def maestral_lock(config_name: str) -> Lock:
    """
    Returns an inter-process and inter-thread lock for Maestral. This is a wrapper
    around :class:`Lock` which fills out the appropriate lockfile path for he given
    config name.

    :param config_name: The name of the Maestral configuration.
    :returns: Lock instance for the config name
    """
    name = f"{config_name}.lock"
    path = get_runtime_path("maestral")
    return Lock.singleton(os.path.join(path, name))


def sockpath_for_config(config_name: str) -> str:
    """
    Returns the unix socket location to be used for the config. This should default to
    the apps runtime directory + 'CONFIG_NAME.sock'.

    :param config_name: The name of the Maestral configuration.
    :returns: Socket path.
    """
    return get_runtime_path("maestral", f"{config_name}.sock")


def lockpath_for_config(config_name: str) -> str:
    """
    Returns the lock file location to be used for the config. This will be the apps
    runtime directory + 'CONFIG_NAME.lock'.

    :param config_name: The name of the Maestral configuration.
    :returns: Path of lock file to use.
    """
    return get_runtime_path("maestral", f"{config_name}.lock")


def get_maestral_pid(config_name: str) -> Optional[int]:
    """
    Returns the PID of the daemon if it is running, ``None`` otherwise.

    :param config_name: The name of the Maestral configuration.
    :returns: The daemon's PID.
    """

    return maestral_lock(config_name).locking_pid()


def is_running(config_name: str) -> bool:
    """
    Checks if a daemon is currently running.

    :param config_name: The name of the Maestral configuration.
    :returns: Whether the daemon is running.
    """

    return maestral_lock(config_name).locked()


def wait_for_startup(config_name: str, timeout: float = 5) -> None:
    """
    Waits until we can communicate with the maestral daemon for ``config_name``.

    :param config_name: Configuration to connect to.
    :param timeout: Timeout it seconds until we raise an error.
    :raises CommunicationError: if we cannot communicate with the daemon within the
        given timeout.
    """

    sock_name = sockpath_for_config(config_name)
    maestral_daemon = Proxy(URI.format(config_name, "./u:" + sock_name))

    t0 = time.time()

    while True:
        try:
            maestral_daemon._pyroBind()
            return
        except Exception as exc:
            if time.time() - t0 > timeout:
                raise exc
            else:
                time.sleep(0.2)
        finally:
            maestral_daemon._pyroRelease()


# ==== main functions to manage daemon =================================================


def start_maestral_daemon(
    config_name: str = "maestral", log_to_stderr: bool = False, start_sync: bool = False
) -> None:
    """
    Starts the Maestral daemon with event loop in the current thread.

    Startup is race free: there will never be more than one daemon running with the same
    config name. The daemon is a :class:`maestral.main.Maestral` instance which is
    exposed as Pyro daemon object and listens for requests on a unix domain socket. This
    call starts an asyncio event loop to process client requests and blocks until the
    event loop shuts down. On macOS, the event loop is integrated with Cocoa's
    CFRunLoop. This allows processing Cocoa events and callbacks, for instance for
    desktop notifications.

    :param config_name: The name of the Maestral configuration to use.
    :param log_to_stderr: If ``True``, write logs to stderr.
    :param start_sync: If ``True``, start syncing once the daemon has started. If the
        ``start_sync`` call fails, an error will be logged but not raised.
    :raises RuntimeError: if a daemon for the given ``config_name`` is already running.
    """

    import asyncio
    from .main import Maestral
    from .logging import scoped_logger, setup_logging

    setup_logging(config_name, log_to_stderr)
    dlogger = scoped_logger(__name__, config_name)

    dlogger.info("Starting daemon")

    if threading.current_thread() is not threading.main_thread():
        dlogger.error("Must run daemon in main thread")
        raise RuntimeError("Must run daemon in main thread")

    dlogger.debug("Environment:\n%s", pformat(os.environ.copy()))

    # acquire PID lock file
    lock = maestral_lock(config_name)

    if lock.acquire():
        dlogger.debug("Acquired daemon lock: %s", lock.path)
    else:
        dlogger.error("Could not acquire lock, daemon is already running")
        raise RuntimeError("Daemon is already running")

    # Nice ourselves to give other processes priority.
    os.nice(10)

    # Integrate with CFRunLoop in macOS.
    if IS_MACOS:

        dlogger.debug("Integrating with CFEventLoop")

        from rubicon.objc.eventloop import EventLoopPolicy  # type: ignore

        asyncio.set_event_loop_policy(EventLoopPolicy())

    # Get the default event loop.
    loop = asyncio.get_event_loop()

    sd_notifier = sdnotify.SystemdNotifier()

    # Notify systemd that we have started.
    if NOTIFY_SOCKET:
        dlogger.debug("Running as systemd notify service")
        dlogger.debug("NOTIFY_SOCKET = %s", NOTIFY_SOCKET)
        sd_notifier.notify("READY=1")

    # Notify systemd periodically if alive.
    if IS_WATCHDOG and WATCHDOG_USEC:

        async def periodic_watchdog() -> None:

            if WATCHDOG_USEC:

                sleep = int(WATCHDOG_USEC)
                while True:
                    sd_notifier.notify("WATCHDOG=1")
                    await asyncio.sleep(sleep / (2 * 10 ** 6))

        dlogger.debug("Running as systemd watchdog service")
        dlogger.debug("WATCHDOG_USEC = %s", WATCHDOG_USEC)
        dlogger.debug("WATCHDOG_PID = %s", WATCHDOG_PID)
        loop.create_task(periodic_watchdog())

    # Get socket for config name.
    sockpath = sockpath_for_config(config_name)
    dlogger.debug(f"Socket path: '{sockpath}'")

    # Clean up old socket.
    try:
        os.remove(sockpath)
    except FileNotFoundError:
        pass

    # Expose maestral as Pyro server. Convert management
    # methods to one way calls so that they don't block.

    dlogger.debug("Creating Pyro daemon")

    ExposedMaestral = expose(Maestral)

    ExposedMaestral.start_sync = oneway(ExposedMaestral.start_sync)
    ExposedMaestral.stop_sync = oneway(ExposedMaestral.stop_sync)
    ExposedMaestral.shutdown_daemon = oneway(ExposedMaestral.shutdown_daemon)

    maestral_daemon = ExposedMaestral(config_name, log_to_stderr=log_to_stderr)

    if start_sync:
        dlogger.debug("Starting sync")
        maestral_daemon.start_sync()

    try:

        dlogger.debug("Starting event loop")

        with Daemon(unixsocket=sockpath) as daemon:
            daemon.register(maestral_daemon, f"maestral.{config_name}")

            # Reduce Pyro's housekeeping frequency from 2 sec to 20 sec.
            # This avoids constantly waking the CPU when we are idle.
            if daemon.transportServer.housekeeper:
                daemon.transportServer.housekeeper.waittime = 20

            for socket in daemon.sockets:
                loop.add_reader(socket.fileno(), daemon.events, daemon.sockets)

            # Handle sigterm gracefully.
            signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
            for s in signals:
                loop.add_signal_handler(s, maestral_daemon.shutdown_daemon)

            loop.run_until_complete(maestral_daemon.shutdown_complete)

            for socket in daemon.sockets:
                loop.remove_reader(socket.fileno())

            # Prevent Pyro housekeeping from blocking shutdown.
            daemon.transportServer.housekeeper = None

    except Exception as exc:
        dlogger.error(exc.args[0], exc_info=True)
    finally:

        if NOTIFY_SOCKET:
            # Notify systemd that we are shutting down.
            sd_notifier.notify("STOPPING=1")


def start_maestral_daemon_process(
    config_name: str = "maestral",
    start_sync: bool = False,
    timeout: int = 5,
) -> Start:
    """
    Starts the Maestral daemon in a new process by calling :func:`start_maestral_daemon`.

    Startup is race free: there will never be more than one daemon running for the same
    config name. This function will use :obj:`sys.executable` as a Python executable to
    start the daemon.

    Environment variables from the current process will be preserved and updated with
    the environment variables defined in :const:`constants.ENV`.

    :param config_name: The name of the Maestral configuration to use.
    :param start_sync: If ``True``, start syncing once the daemon has started.
    :param timeout: Time in sec to wait for daemon to start.
    :returns: :attr:`Start.Ok` if successful, :attr:`Start.AlreadyRunning` if the daemon
        was already running or :attr:`Start.Failed` if startup failed. It is possible
        that :attr:`Start.Ok` may be returned instead of :attr:`Start.AlreadyRunning`
        in case of a race but the daemon is nevertheless started only once.
    """

    if is_running(config_name):
        return Start.AlreadyRunning

    # Protect against injection.
    cc = quote(config_name).strip("'")
    start_sync = bool(start_sync)

    script = (
        f"import maestral.daemon; "
        f'maestral.daemon.start_maestral_daemon("{cc}", start_sync={start_sync})'
    )

    cmd = [sys.executable, "-c", script]

    subprocess_env = os.environ.copy()
    subprocess_env.update(ENV)

    process = subprocess.Popen(
        cmd,
        start_new_session=True,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        env=subprocess_env,
    )

    try:
        wait_for_startup(config_name, timeout)
    except Exception as exc:
        from .logging import scoped_logger, setup_logging

        setup_logging(config_name, log_to_stderr=False)
        clogger = scoped_logger(__name__, config_name)

        clogger.error("Could not communicate with daemon", exc_info=exc_info_tuple(exc))

        # Let's check what the daemon has been doing.
        returncode = process.poll()
        if returncode is None:
            clogger.error("Daemon is running but not responsive, killing now")
            process.terminate()
        else:
            clogger.error("Daemon stopped with return code %s", returncode)
        return Start.Failed
    else:
        return Start.Ok


def stop_maestral_daemon_process(
    config_name: str = "maestral", timeout: float = 10
) -> Stop:
    """Stops a maestral daemon process by finding its PID and shutting it down.

    This function first tries to shut down Maestral gracefully. If this fails and we
    know its PID, it will send SIGTERM. If that fails as well, it will send SIGKILL to
    the process.

    :param config_name: The name of the Maestral configuration to use.
    :param timeout: Number of sec to wait for daemon to shut down before killing it.
    :returns: :attr:`Stop.Ok` if successful, :attr:`Stop.Killed` if killed,
        :attr:`Stop.NotRunning` if the daemon was not running and :attr:`Stop.Failed`
        if killing the process failed because we could not retrieve its PID.
    """

    if not is_running(config_name):
        return Stop.NotRunning

    pid = get_maestral_pid(config_name)

    try:
        with MaestralProxy(config_name) as m:
            m.shutdown_daemon()
    except CommunicationError:
        if pid:
            _send_term(pid)
    finally:
        while timeout > 0:
            if not is_running(config_name):
                return Stop.Ok
            else:
                time.sleep(0.2)
                timeout -= 0.2

        # Send SIGTERM after timeout and delete PID file.
        if pid:
            _send_term(pid)

        time.sleep(1)

        if not is_running(config_name):
            return Stop.Ok
        elif pid:
            os.kill(pid, signal.SIGKILL)
            return Stop.Killed
        else:
            return Stop.Failed


class MaestralProxy:
    """A Proxy to the Maestral daemon

    All methods and properties of Maestral's public API are accessible and calls /
    access will be forwarded to the corresponding Maestral instance. This class can be
    used as a context manager to close the connection to the daemon on exit.

    :Example:

        Use MaestralProxy as a context manager:

        >>> with MaestralProxy() as m:
        ...     print(m.status)

        Use MaestralProxy directly:

        >>> m = MaestralProxy()
        >>> print(m.status)
        >>> m._disconnect()

    :ivar _is_fallback: Whether we are using an actual Maestral instance as fallback
        instead of a Proxy.

    :param config_name: The name of the Maestral configuration to use.
    :param fallback: If ``True``, a new instance of Maestral will created in the current
        process when the daemon is not running.
    :raises CommunicationError: if the daemon is running but cannot be reached or if the
        daemon is not running and ``fallback`` is ``False``.
    """

    _m: Union["Maestral", Proxy]

    def __init__(self, config_name: str = "maestral", fallback: bool = False) -> None:

        self._config_name = config_name
        self._is_fallback = False

        if is_running(config_name):

            sock_name = sockpath_for_config(config_name)

            # print remote tracebacks locally
            sys.excepthook = Pyro5.errors.excepthook

            self._m = Proxy(URI.format(config_name, "./u:" + sock_name))
            try:
                self._m._pyroBind()
            except CommunicationError:
                self._m._pyroRelease()
                raise

        else:
            # If daemon is not running, fall back to new Maestral instance
            # or raise a CommunicationError if fallback not allowed.
            if fallback:
                from .main import Maestral

                self._m = Maestral(config_name)
            else:
                raise CommunicationError(f"Could not get proxy for '{config_name}'")

        self._is_fallback = not isinstance(self._m, Proxy)

    def _disconnect(self) -> None:
        if isinstance(self._m, Proxy):
            self._m._pyroRelease()

    def __enter__(self) -> "MaestralProxy":
        return self

    def __exit__(
        self, exc_type: Type[Exception], exc_value: Exception, tb: TracebackType
    ) -> None:
        self._disconnect()
        del self._m

    def __getattr__(self, item: str) -> Any:
        if item.startswith("_"):
            super().__getattribute__(item)
        elif isinstance(self._m, Proxy):
            return self._m.__getattr__(item)
        else:
            return self._m.__getattribute__(item)

    def __setattr__(self, key, value) -> None:
        if key.startswith("_"):
            super().__setattr__(key, value)
        else:
            self._m.__setattr__(key, value)

    def __dir__(self) -> Iterable[str]:
        own_result = dir(self.__class__) + list(self.__dict__.keys())
        proxy_result = [k for k in self._m.__dir__() if not k.startswith("_")]

        return sorted(set(own_result) | set(proxy_result))

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}(config={self._config_name!r}, "
            f"is_fallback={self._is_fallback})>"
        )


def get_maestral_proxy(
    config_name: str = "maestral", fallback: bool = False
) -> Union["Maestral", Proxy]:

    warnings.warn(
        "'get_maestral_proxy' is deprecated, please use 'MaestralProxy' instead",
        DeprecationWarning,
    )

    m = MaestralProxy(config_name, fallback=fallback)
    return m._m



# -*- coding: utf-8 -*-

import warnings

from click import shell_completion  # type: ignore


__version__ = "1.4.5.dev0"
__author__ = "Sam Schott"
__url__ = "https://maestral.app"


# suppress Python 3.9 warning from rubicon-objc
warnings.filterwarnings("ignore", module="rubicon", category=UserWarning)


# patch click shell completion argument detection
# see https://github.com/pallets/click/issues/1929


def _start_of_option(value: str) -> bool:
    """Check if the value looks like the start of an option."""
    return value[0] == "-" if value else False


shell_completion._start_of_option = _start_of_option



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
This module provides user configuration file management and is mostly copied from the
config module of the Spyder IDE.
"""

import ast
import os
import os.path as osp
import re
import shutil
import time
import copy
import configparser as cp
from threading import RLock
import logging
from typing import Optional, Dict, Any


logger = logging.getLogger(__name__)

DefaultsType = Dict[str, Dict[str, Any]]

# =============================================================================
# Auxiliary classes
# =============================================================================


class NoDefault:
    pass


# =============================================================================
# Defaults class
# =============================================================================


class DefaultsConfig(cp.ConfigParser):
    """
    Class used to save defaults to a file and as base class for UserConfig.
    """

    def __init__(self, path: str) -> None:
        super().__init__(interpolation=None)

        dirname, basename = osp.split(path)
        filename, ext = osp.splitext(basename)

        self._path = path
        self._dirname = dirname
        self._filename = filename
        self._suffix = ext

    def _set(self, section: str, option: str, value: Any) -> None:
        """Private set method"""
        if not self.has_section(section):
            self.add_section(section)
        if not isinstance(value, str):
            value = repr(value)

        super().set(section, option, value)

    def save(self) -> None:
        """Save config into the associated file."""
        fpath = self.get_config_fpath()

        # See spyder-ide/spyder#1086 and spyder-ide/spyder#1242 for background
        # on why this method contains all the exception handling.

        try:
            # The "easy" way
            self.__write_file(fpath)
        except EnvironmentError:
            try:
                # The "delete and sleep" way
                if osp.isfile(fpath):
                    os.remove(fpath)

                time.sleep(0.05)
                self.__write_file(fpath)
            except Exception:
                logger.warning(
                    "Failed to write user configuration to disk", exc_info=True
                )

    def __write_file(self, fpath: str) -> None:

        os.makedirs(self._dirname, exist_ok=True)

        with open(fpath, "w", encoding="utf-8") as configfile:
            self.write(configfile)

    def get_config_fpath(self) -> str:
        """Return the ini file where this configuration is stored."""
        return self._path


# =============================================================================
# User config class
# =============================================================================


class UserConfig(DefaultsConfig):
    """
    UserConfig class, based on ConfigParser.

    Parameters
    ----------
    path:
        Configuration file will be saved to this path.
    defaults:
        Dictionary containing options *or* list of tuples (sec_name, options)
    load:
        If a previous configuration file is found, load will take the values
        from this existing file, instead of using default values.
    version:
        version of the configuration file in 'major.minor.micro' format.
    backup:
        A backup will be created on version changes and on initial setup.
    remove_obsolete:
        If `True`, values that were removed from the configuration on version
        change, are removed from the saved configuration file.

    Notes
    -----
    The 'get' and 'set' arguments number and type differ from the reimplemented
    methods. 'defaults' is an attribute and not a method.
    """

    DEFAULT_SECTION_NAME = "main"

    def __init__(
        self,
        path: str,
        defaults: Optional[DefaultsType] = None,
        load: bool = True,
        version: str = "0.0.0",
        backup: bool = False,
        remove_obsolete: bool = False,
    ) -> None:
        """UserConfig class, based on ConfigParser."""
        super().__init__(path=path)

        self._lock = RLock()

        self._load = load
        self._backup = backup
        self._remove_obsolete = remove_obsolete

        self._version = self._check_version(version)
        self.default_config = self._set_defaults(defaults)

        # Set all values to defaults. They may be overwritten later
        # when loading form file.
        self.reset_to_defaults(save=False)

        self._backup_folder = "backups"
        self._backup_suffix = ".bak"

        if backup:
            self._make_backup()

        if load:
            # If config file already exists, it overrides Default options.
            self._load_from_ini(self.get_config_fpath())
            old_version = self.get_version(version)

            # Updating defaults only if major/minor version is different.
            major_ver = self._get_major_version(version)
            major_old_ver = self._get_major_version(old_version)

            minor_ver = self._get_minor_version(version)
            minor_old_ver = self._get_minor_version(old_version)

            if major_ver != major_old_ver or minor_ver != minor_old_ver:

                if backup:
                    self._make_backup(version=old_version)

                self.apply_configuration_patches(old_version)

                # Remove deprecated options if major version has changed.
                if remove_obsolete and major_ver != major_old_ver:
                    self._remove_deprecated_options()

                # Set new version number.
                self.set_version(version, save=False)

            # Save any changes back to file.
            self.save()

    # --- Helpers and checkers ---------------------------------------------------------

    @staticmethod
    def _get_minor_version(version: str) -> str:
        """Return the 'major.minor' components of the version."""
        return version[: version.rfind(".")]

    @staticmethod
    def _get_major_version(version: str) -> str:
        """Return the 'major' component of the version."""
        return version[: version.find(".")]

    @staticmethod
    def _check_version(version: str) -> str:
        """Check version is compliant with format."""
        regex_check = re.match(r"^(\d+).(\d+).(\d+)$", version)
        if regex_check is None:
            raise ValueError(
                "Version number {} is incorrect - must be in "
                "major.minor.micro format".format(version)
            )

        return version

    def _set_defaults(self, defaults: Optional[DefaultsType]) -> DefaultsType:
        """Check if defaults are valid and update defaults values."""

        defaults = copy.deepcopy(defaults)

        if not defaults:
            defaults = {}

        if self.DEFAULT_SECTION_NAME not in defaults:
            defaults[self.DEFAULT_SECTION_NAME] = {}

        self.default_config = defaults
        self.default_config[self.DEFAULT_SECTION_NAME]["version"] = self._version

        return self.default_config

    def _check_section(self, section: Optional[str]) -> str:
        """Check section."""
        return section or self.DEFAULT_SECTION_NAME

    def _make_backup(
        self, version: Optional[str] = None, old_version: Optional[str] = None
    ) -> None:
        """
        Make a backup of the configuration file.

        If `old_version` is `None` a normal backup is made. If `old_version`
        is provided, then the backup was requested for minor version changes
        and appends the version number to the backup file.
        """
        fpath = self.get_config_fpath()
        fpath_backup = self.get_backup_fpath_from_version(version=version)
        path = os.path.dirname(fpath_backup)

        if not osp.isdir(path):
            os.makedirs(path)

        try:
            shutil.copyfile(fpath, fpath_backup)
        except IOError:
            pass

    def _load_from_ini(self, fpath: str) -> None:
        """Load config from the associated file found at `fpath`."""

        with self._lock:
            try:
                self.read(fpath, encoding="utf-8")
            except cp.MissingSectionHeaderError:
                logger.error("File contains no section headers.")

    def _remove_deprecated_options(self) -> None:
        """
        Remove options which are present in the file but not in defaults.
        """
        for section in self.sections():
            for option, _ in self.items(section, raw=True):
                if self.get_default(section, option) is NoDefault:
                    try:
                        self.remove_option(section, option)
                        if len(self.items(section, raw=True)) == 0:
                            self.remove_section(section)
                    except cp.NoSectionError:
                        self.remove_section(section)

    # --- Compatibility API ------------------------------------------------------------

    def get_backup_fpath_from_version(self, version: Optional[str] = None) -> str:
        """
        Get backup location based on version.
        """
        path = osp.join(self._dirname, self._backup_folder)

        if version is None:
            name = f"{self._filename}{self._backup_suffix}"
        else:
            name = f"{self._filename}-{version}{self._backup_suffix}"

        return osp.join(path, name)

    def apply_configuration_patches(self, old_version: str = None) -> None:
        """
        Apply any patch to configuration values on version changes.

        To be reimplemented if patches to configuration values are needed.
        """
        pass

    # --- Public API -------------------------------------------------------------------

    def get_version(self, version: str = "0.0.0") -> str:
        """Return configuration (not application!) version."""
        with self._lock:
            return self.get(self.DEFAULT_SECTION_NAME, "version", version)

    def set_version(self, version: str = "0.0.0", save: bool = True) -> None:
        """Set configuration (not application!) version."""

        with self._lock:
            version = self._check_version(version)
            self.set(self.DEFAULT_SECTION_NAME, "version", version, save=save)

    def reset_to_defaults(
        self, save: bool = True, section: Optional[str] = None
    ) -> None:
        """Reset config to Default values."""

        with self._lock:

            for sec, options in self.default_config.items():
                if section is None or section == sec:
                    for option in options:
                        value = options[option]
                        self._set(sec, option, value)

            if save:
                self.save()

    def get_default(self, section: str, option: str) -> Any:
        """
        Get default value for a given `section` and `option`.

        This is useful for type checking in `get` method.
        """

        with self._lock:
            secdict = self.default_config.get(section, {})
            return secdict.get(option, NoDefault)

    def get(self, section: str, option: str, default: Any = NoDefault) -> Any:  # type: ignore
        """
        Get an option.

        Parameters
        ----------
        section:
            Section name. If `None` is provide use the default section name.
        option:
            Option name for `section`.
        default:
            Default value (if not specified, an exception will be raised if
            option doesn't exist).
        """

        with self._lock:

            section = self._check_section(section)

            if not self.has_section(section):
                if default is NoDefault:
                    raise cp.NoSectionError(section)
                else:
                    self.add_section(section)

            if not self.has_option(section, option):
                if default is NoDefault:
                    raise cp.NoOptionError(option, section)
                else:
                    self.set(section, option, default)
                    return default

            raw_value: str = super(UserConfig, self).get(section, option, raw=True)
            default_value = self.get_default(section, option)
            value: Any

            if isinstance(default_value, str):
                value = raw_value
            else:
                try:
                    value = ast.literal_eval(raw_value)
                except (SyntaxError, ValueError):
                    value = raw_value

            if default_value is not NoDefault and type(default_value) is not type(
                value
            ):
                logger.error(
                    f"Inconsistent config type for [{section}][{option}]. "
                    f"Expected {default_value.__class__.__name__} but "
                    f"got {value.__class__.__name__}."
                )

            return value

    def set_default(self, section: str, option: str, default_value: Any) -> None:
        """
        Set Default value for a given `section`, `option`.

        If the section or option does not exist, it will be created.
        """

        with self._lock:

            if section not in self.default_config:
                self.default_config[section] = {}

            self.default_config[section][option] = default_value

    def set(self, section: str, option: str, value: Any, save: bool = True) -> None:  # type: ignore
        """
        Set an `option` on a given `section`.

        If section is None, the `option` is added to the default section.
        """

        with self._lock:

            section = self._check_section(section)
            default_value = self.get_default(section, option)

            if default_value is NoDefault:
                default_value = value
                self.set_default(section, option, default_value)

            if isinstance(default_value, float) and isinstance(value, int):
                value = float(value)

            if type(default_value) is not type(value):
                raise ValueError(
                    f"Inconsistent config type for [{section}][{option}]. "
                    f"Expected {default_value.__class__.__name__} but "
                    f"got {value.__class__.__name__}."
                )

            self._set(section, option, value)
            if save:
                self.save()

    def remove_section(self, section: str) -> bool:
        """Remove `section` and all options within it."""

        with self._lock:
            res = super().remove_section(section)
            self.save()
            return res

    def remove_option(self, section: str, option: str) -> bool:
        """Remove `option` from `section`."""

        with self._lock:
            res = super().remove_option(section, option)
            self.save()
            return res

    def cleanup(self) -> None:
        """Remove files associated with config and reset to defaults."""

        with self._lock:

            self.reset_to_defaults(save=False)
            backup_path = osp.join(self._dirname, self._backup_folder)

            # remove config file
            try:
                os.remove(self.get_config_fpath())
            except FileNotFoundError:
                pass

            # remove saved backups
            for file in os.scandir(backup_path):
                if file.name.startswith(self._filename):
                    try:
                        os.remove(file.path)
                    except FileNotFoundError:
                        pass



# -*- coding: utf-8 -*-
"""This module defines custom logging records and handlers."""

import os
import concurrent.futures
import logging
from logging.handlers import RotatingFileHandler
from collections import deque
from concurrent.futures import Future
from typing import Deque, Optional, List, Tuple, Union

try:
    from concurrent.futures import InvalidStateError  # type: ignore
except ImportError:
    # Python 3.7 and lower
    InvalidStateError = RuntimeError  # type: ignore

import sdnotify  # type: ignore

try:
    from systemd import journal  # type: ignore
except ImportError:
    journal = None

from .config import MaestralConfig
from .utils import sanitize_string
from .utils.appdirs import get_log_path


__all__ = [
    "CachedHandler",
    "SdNotificationHandler",
    "setup_logging",
    "scoped_logger",
]


def safe_journal_sender(MESSAGE: str, **kwargs) -> None:

    if journal:

        MESSAGE = sanitize_string(MESSAGE)

        for key, value in kwargs.items():
            if isinstance(value, str):
                kwargs[key] = sanitize_string(value)

        journal.send(MESSAGE, **kwargs)


class EncodingSafeLogRecord(logging.LogRecord):
    """A log record which ensures that messages contain only unicode characters

    This is useful when log messages may contain file paths generates by OS APIs. In
    Python, such path strings may contain surrogate escapes and will therefore raise
    a :class:`UnicodeEncodeError` under many circumstances (printing to stdout, etc).
    """

    _msg: Optional[str] = None

    def getMessage(self) -> str:
        """
        Formats the log message and replaces all surrogate escapes with "�".
        """
        if not self._msg:
            self._msg = sanitize_string(super().getMessage())

        return self._msg


logging.setLogRecordFactory(EncodingSafeLogRecord)


class CachedHandler(logging.Handler):
    """Handler which stores past records

    This is used to populate Maestral's status and error interfaces. The method
    :meth:`wait_for_emit` can be used from another thread to block until a new record is
    emitted, for instance to react to state changes.

    :param level: Initial log level. Defaults to NOTSET.
    :param maxlen: Maximum number of records to store. If ``None``, all records will be
        stored. Defaults to ``None``.
    """

    cached_records: Deque[logging.LogRecord]
    _emit_future: Future

    def __init__(
        self, level: int = logging.NOTSET, maxlen: Optional[int] = None
    ) -> None:
        super().__init__(level=level)
        self.cached_records = deque([], maxlen)
        self._emit_future = Future()

    def emit(self, record: logging.LogRecord) -> None:
        """
        Logs the specified log record and saves it to the cache.

        :param record: Log record.
        """
        self.cached_records.append(record)

        # notify any waiting coroutines that we have a status change
        try:
            self._emit_future.set_result(True)
        except InvalidStateError:
            pass

    def wait_for_emit(self, timeout: Optional[float]) -> bool:
        """
        Blocks until a new record is emitted.

        :param timeout: Maximum time to block before returning.
        :returns: ``True`` if there was a status change, ``False`` in case of a timeout.
        """
        try:
            self._emit_future.result(timeout=timeout)
        except concurrent.futures.TimeoutError:
            return False

        self._emit_future = Future()  # reset future
        return True

    def getLastMessage(self) -> str:
        """
        :returns: The log message of the last record or an empty string.
        """
        try:
            last_record = self.cached_records[-1]
            return last_record.getMessage()
        except IndexError:
            return ""

    def getAllMessages(self) -> List[str]:
        """
        :returns: A list of all record messages.
        """
        return [r.getMessage() for r in self.cached_records]

    def clear(self) -> None:
        """
        Clears all cached records.
        """
        self.cached_records.clear()


class SdNotificationHandler(logging.Handler):
    """Handler which emits messages as systemd notifications

    This is useful when used from a systemd service and will do nothing when no
    NOTIFY_SOCKET is provided.
    """

    notifier = sdnotify.SystemdNotifier()

    def emit(self, record: logging.LogRecord) -> None:
        """
        Sends the record massage to systemd as service status.

        :param record: Log record.
        """
        self.notifier.notify(f"STATUS={record.message}")


def scoped_logger_name(module_name: str, config_name: str = "maestral") -> str:
    """
    Returns a logger name for the module ``module_name``, scoped to the given config.

    :param module_name: Module name.
    :param config_name: Config name.
    :returns: Scoped logger name.
    """

    if config_name == "maestral":
        return module_name
    else:
        return f"{config_name}-{module_name}"


def scoped_logger(module_name: str, config_name: str = "maestral") -> logging.Logger:
    """
    Returns a logger for the module ``module_name``, scoped to the given config.

    :param module_name: Module name.
    :param config_name: Config name.
    :returns: Logger instances scoped to the config.
    """

    return logging.getLogger(scoped_logger_name(module_name, config_name))


def setup_logging(
    config_name: str, log_to_stderr: bool = True
) -> Tuple[
    RotatingFileHandler,
    Union[logging.StreamHandler, logging.NullHandler],
    SdNotificationHandler,
    Union["journal.JournalHandler", logging.NullHandler],
]:
    """
    Sets up logging handlers for the given config name. The following handlers are
    installed for the root logger:

    * RotatingFileHandler: Writes logs to the appropriate log file for the config. Log
      level is determined by the config value.
    * StreamHandler: Writes logs to stderr. Log level is determined by the config value.
      This will be replaced by a null handler if ``log_to_stderr`` is ``False``.
    * SdNotificationHandler: Sends all log messages of level INFO and higher to the
      NOTIFY_SOCKET if provided as an environment variable. The log level is fixed.
    * JournalHandler: Writes logs to the systemd journal. Log level is determined by the
      config value. Will be replaced by a null handler if not started as a systemd
      service or if python-systemd is not installed.

    Any previous loggers are cleared.

    :param config_name: The config name.
    :param log_to_stderr: Whether to log to stderr.
    :returns: (log_handler_file, log_handler_stream, log_handler_sd, log_handler_journal)
    """

    conf = MaestralConfig(config_name)

    # Get log level from config or fallback to DEBUG level if config file is corrupt.
    log_level = conf.get("app", "log_level", logging.DEBUG)

    root_logger = scoped_logger("maestral", config_name)
    root_logger.setLevel(min(log_level, logging.INFO))

    root_logger.handlers = []  # clean up any previous handlers

    log_fmt_long = logging.Formatter(
        fmt="%(asctime)s %(module)s %(levelname)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    log_fmt_short = logging.Formatter(fmt="%(message)s")

    # Log to file.
    log_file_path = get_log_path("maestral", f"{config_name}.log")
    log_handler_file = RotatingFileHandler(
        log_file_path, maxBytes=10 ** 7, backupCount=1
    )
    log_handler_file.setFormatter(log_fmt_long)
    log_handler_file.setLevel(log_level)
    root_logger.addHandler(log_handler_file)

    # Log to systemd journal when running as systemd service.
    log_handler_journal: Union["journal.JournalHandler", logging.NullHandler]

    if journal and os.getenv("INVOCATION_ID"):
        log_handler_journal = journal.JournalHandler(
            SYSLOG_IDENTIFIER="maestral",
            sender_function=safe_journal_sender,
        )
    else:
        log_handler_journal = logging.NullHandler()

    log_handler_journal.setFormatter(log_fmt_short)
    log_handler_journal.setLevel(log_level)
    root_logger.addHandler(log_handler_journal)

    # Log to NOTIFY_SOCKET when launched as systemd notify service.
    log_handler_sd = SdNotificationHandler()
    log_handler_sd.setFormatter(log_fmt_short)
    log_handler_sd.setLevel(logging.INFO)
    root_logger.addHandler(log_handler_sd)

    # Log to stderr if requested.

    log_handler_stream: Union[logging.StreamHandler, logging.NullHandler]

    if log_to_stderr:
        log_handler_stream = logging.StreamHandler()
    else:
        log_handler_stream = logging.NullHandler()
    log_handler_stream.setFormatter(log_fmt_long)
    log_handler_stream.setLevel(log_level)
    root_logger.addHandler(log_handler_stream)

    return log_handler_file, log_handler_stream, log_handler_sd, log_handler_journal



# -*- coding: utf-8 -*-
"""
This module defines the functions to configure and interact with Maestral from the
command line. Some imports are deferred to the functions that required them in order to
reduce the startup time of individual CLI commands.
"""

# system imports
import sys
import os
import os.path as osp
import functools
import time
from typing import Optional, Dict, List, Tuple, Callable, Union, cast, TYPE_CHECKING

# external imports
import click

# local imports
from . import __version__
from .utils import cli

if TYPE_CHECKING:
    from click.shell_completion import CompletionItem
    from datetime import datetime
    from .main import Maestral
    from .daemon import MaestralProxy


# ======================================================================================
# CLI dialogs and helper functions
# ======================================================================================

OK = click.style("[OK]", fg="green")
FAILED = click.style("[FAILED]", fg="red")
KILLED = click.style("[KILLED]", fg="red")


def stop_daemon_with_cli_feedback(config_name: str) -> None:
    """Wrapper around :meth:`daemon.stop_maestral_daemon_process`
    with command line feedback."""

    from .daemon import stop_maestral_daemon_process, Stop

    click.echo("Stopping Maestral...", nl=False)
    res = stop_maestral_daemon_process(config_name)
    if res == Stop.Ok:
        click.echo("\rStopping Maestral...        " + OK)
    elif res == Stop.NotRunning:
        click.echo("\rMaestral daemon is not running.")
    elif res == Stop.Killed:
        click.echo("\rStopping Maestral...        " + KILLED)
    elif res == Stop.Failed:
        click.echo("\rStopping Maestral...        " + FAILED)


def select_dbx_path_dialog(
    config_name: str, default_dir_name: Optional[str] = None, allow_merge: bool = False
) -> str:
    """
    A CLI dialog to ask for a local Dropbox folder location.

    :param config_name: The configuration to use for the default folder name.
    :param default_dir_name: The default directory name. Defaults to
        "Dropbox ({config_name})" if not given.
    :param allow_merge: If ``True``, allows the selection of an existing folder without
        deleting it. Defaults to ``False``.
    :returns: Path given by user.
    """

    from .utils.path import delete

    default_dir_name = default_dir_name or f"Dropbox ({config_name.capitalize()})"

    while True:
        res = cli.select_path(
            "Please choose a local Dropbox folder:",
            default=f"~/{default_dir_name}",
            files_allowed=False,
        )
        res = res.rstrip(osp.sep)

        dropbox_path = osp.expanduser(res)

        if osp.exists(dropbox_path):
            if allow_merge:
                text = (
                    "Directory already exists. Do you want to replace it "
                    "or merge its content with your Dropbox?"
                )
                choice = cli.select(text, options=["replace", "merge", "cancel"])
            else:
                text = (
                    "Directory already exists. Do you want to replace it? "
                    "Its content will be lost!"
                )
                replace = cli.confirm(text)
                choice = 0 if replace else 2

            if choice == 0:
                err = delete(dropbox_path)
                if err:
                    cli.warn(
                        "Could not write to selected location. "
                        "Please make sure that you have sufficient permissions."
                    )
                else:
                    cli.ok("Replaced existing folder")
                    return dropbox_path
            elif choice == 1:
                cli.ok("Merging with existing folder")
                return dropbox_path

        else:
            return dropbox_path


def link_dialog(m: Union["MaestralProxy", "Maestral"]) -> None:
    """
    A CLI dialog for linking a Dropbox account.

    :param m: Proxy to Maestral daemon.
    """

    authorize_url = m.get_auth_url()

    cli.info(f"Linking new account for '{m.config_name}' config")
    cli.info("Retrieving auth code from Dropbox")
    choice = cli.select(
        "How would you like to you link your account?",
        options=["Open Dropbox website", "Print auth URL to console"],
    )

    if choice == 0:
        click.launch(authorize_url)
    else:
        cli.info("Open the URL below to retrieve an auth code:")
        cli.info(authorize_url)

    res = -1
    while res != 0:
        auth_code = cli.prompt("Enter the auth code:")
        auth_code = auth_code.strip()

        res = m.link(auth_code)

        if res == 0:
            email = m.get_state("account", "email")
            cli.ok(f"Linked to {email}")
        elif res == 1:
            cli.warn("Invalid token, please try again")
        elif res == 2:
            cli.warn("Could not connect to Dropbox, please try again")


def check_for_updates() -> None:
    """
    Checks if updates are available by reading the cached release number from the
    config file and notifies the user. Prints an update note to the command line.
    """
    from packaging.version import Version
    from .config import MaestralConfig, MaestralState

    conf = MaestralConfig("maestral")
    state = MaestralState("maestral")

    interval = conf.get("app", "update_notification_interval")
    last_update_check = state.get("app", "update_notification_last")
    latest_release = state.get("app", "latest_release")

    if interval == 0 or time.time() - last_update_check < interval:
        return

    has_update = Version(__version__) < Version(latest_release)

    if has_update:
        cli.echo(
            f"Update available v{__version__} → v{latest_release}. "
            f"Please use your package manager to update."
        )


def check_for_fatal_errors(m: Union["MaestralProxy", "Maestral"]) -> bool:
    """
    Checks the given Maestral instance for fatal errors such as revoked Dropbox access,
    deleted Dropbox folder etc. Prints a nice representation to the command line.

    :param m: Proxy to Maestral daemon or Maestral instance.
    :returns: True in case of fatal errors, False otherwise.
    """

    import textwrap
    import shutil

    maestral_err_list = m.fatal_errors

    if len(maestral_err_list) > 0:

        width, height = shutil.get_terminal_size()

        err = maestral_err_list[0]
        err_title = cast(str, err["title"])
        err_msg = cast(str, err["message"])

        wrapped_msg = textwrap.fill(err_msg, width=width)

        click.echo("")
        click.secho(err_title, fg="red")
        click.secho(wrapped_msg, fg="red")
        click.echo("")

        return True
    else:
        return False


def convert_py_errors(func: Callable) -> Callable:
    """
    Decorator that catches a MaestralApiError and prints a formatted error message to
    stdout before exiting. Calls ``sys.exit(1)`` after printing the error to stdout.
    """

    from .errors import MaestralApiError

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except MaestralApiError as exc:
            cli.warn(f"{exc.title}. {exc.message}")
            sys.exit(1)

    return wrapper


def _datetime_from_iso_str(time_str: str) -> "datetime":
    """
    Converts an ISO 8601 time string such as '2015-05-15T15:50:38Z' to a timezone aware
    datetime object in the local time zone.
    """

    from datetime import datetime

    # replace Z with +0000, required for Python 3.6 compatibility
    time_str = time_str.replace("Z", "+0000")
    return datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%S%z").astimezone()


# ======================================================================================
# Custom parameter types
# ======================================================================================

# A custom parameter:
# * needs a name
# * needs to pass through None unchanged
# * needs to convert from a string
# * needs to convert its result type through unchanged (eg: needs to be idempotent)
# * needs to be able to deal with param and context being None. This can be the case
#   when the object is used with prompt inputs.


class DropboxPath(click.ParamType):
    """A command line parameter representing a Dropbox path

    :param file_okay: Controls if a file is a possible value.
    :param dir_okay: Controls if a directory is a possible value.
    """

    name = "Dropbox path"
    envvar_list_splitter = osp.pathsep

    def __init__(self, file_okay: bool = True, dir_okay: bool = True) -> None:
        self.file_okay = file_okay
        self.dir_okay = dir_okay

    def convert(
        self,
        value: Optional[str],
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> Optional[str]:

        if value is None:
            return value

        if not value.startswith("/"):
            value = "/" + value

        return value

    def shell_complete(
        self,
        ctx: Optional[click.Context],
        param: Optional[click.Parameter],
        incomplete: str,
    ) -> List["CompletionItem"]:

        from click.shell_completion import CompletionItem
        from .utils import removeprefix
        from .config import MaestralConfig

        matches: List[str] = []
        completions: List[CompletionItem] = []

        # check if we have been given an absolute path
        absolute = incomplete.startswith("/")
        incomplete = incomplete.lstrip("/")

        # get the Maestral config for which to complete paths
        config_name = ctx.params.get("config_name", "maestral") if ctx else "maestral"

        # get all matching paths in our local Dropbox folder
        # TODO: query from server if not too slow

        config = MaestralConfig(config_name)
        dropbox_dir = config.get("main", "path")
        local_incomplete = osp.join(dropbox_dir, incomplete)
        local_dirname = osp.dirname(local_incomplete)

        try:
            with os.scandir(local_dirname) as it:
                for entry in it:
                    if entry.path.startswith(local_incomplete):
                        if self.file_okay and entry.is_file():
                            dbx_path = removeprefix(entry.path, dropbox_dir)
                            matches.append(dbx_path)
                        if self.dir_okay and entry.is_dir():
                            dbx_path = removeprefix(entry.path, dropbox_dir)
                            matches.append(dbx_path)
        except OSError:
            pass

        # get all matching excluded items

        for dbx_path in config.get("main", "excluded_items"):
            if dbx_path.startswith("/" + incomplete):
                matches.append(dbx_path)

        for match in matches:
            if not absolute:
                match = match.lstrip("/")
            completions.append(CompletionItem(match))

        return completions


class ConfigKey(click.ParamType):
    """A command line parameter representing a config key"""

    name = "key"

    def shell_complete(
        self,
        ctx: Optional[click.Context],
        param: Optional[click.Parameter],
        incomplete: str,
    ) -> List["CompletionItem"]:

        from click.shell_completion import CompletionItem
        from .config.main import KEY_SECTION_MAP as KEYS

        return [CompletionItem(key) for key in KEYS if key.startswith(incomplete)]


class ConfigName(click.ParamType):
    """A command line parameter representing a Dropbox path

    :param existing: If ``True`` require an existing config, otherwise create a new
        config on demand.
    """

    name = "config"

    def __init__(self, existing: bool = True) -> None:
        self.existing = existing

    def convert(
        self,
        value: Optional[str],
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> Optional[str]:

        if value is None:
            return value

        from .config import validate_config_name, list_configs

        if not self.existing:

            # accept all valid config names
            try:
                return validate_config_name(value)
            except ValueError:
                raise cli.CliException(
                    "Configuration name may not contain any whitespace"
                )

        else:

            # accept only existing config names
            if value in list_configs():
                return value
            else:
                raise cli.CliException(
                    f"Configuration '{value}' does not exist. "
                    f"Use 'maestral configs' to list all configurations."
                )

    def shell_complete(
        self,
        ctx: Optional[click.Context],
        param: Optional[click.Parameter],
        incomplete: str,
    ) -> List["CompletionItem"]:

        from click.shell_completion import CompletionItem
        from .config import list_configs

        matches = [conf for conf in list_configs() if conf.startswith(incomplete)]
        return [CompletionItem(m) for m in matches]


# ======================================================================================
# Command groups
# ======================================================================================


class OrderedGroup(click.Group):
    """Click command group with customizable order of help output."""

    def command(self, *args, **kwargs) -> Callable:
        """Behaves the same as :meth:`click.Group.command()` except captures a section
        name for listing command names in help.
        """
        section = kwargs.pop("section", "Commands")

        from click.decorators import command

        def decorator(f):
            cmd = command(*args, **kwargs)(f)
            cmd.section = section
            self.add_command(cmd)

            return cmd

        return decorator

    def group(self, *args, **kwargs) -> Callable:
        """Behaves the same as :meth:`click.Group.group()` except captures a section
        name for listing command names in help.
        """
        section = kwargs.pop("section", "Commands")

        from click.decorators import group

        def decorator(f):
            cmd = group(*args, **kwargs)(f)
            cmd.section = section
            self.add_command(cmd)

            return cmd

        return decorator

    def format_commands(
        self, ctx: click.Context, formatter: click.HelpFormatter
    ) -> None:
        """Extra format methods for multi methods that adds all the commands
        after the options.
        """
        commands = []

        for name in self.commands:
            cmd = self.get_command(ctx, name)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((name, cmd))

        # allow for 3 times the default spacing
        if len(commands) > 0:
            max_len = max(len(name) for name, cmd in commands)
            limit = formatter.width - 6 - max_len  # type: ignore

            sections: Dict[str, List[Tuple[str, click.Command]]] = {}

            # group commands into sections
            for name, cmd in commands:
                try:
                    sections[cmd.section].append((name, cmd))  # type: ignore
                except KeyError:
                    sections[cmd.section] = [(name, cmd)]  # type: ignore

            # format sections individually
            for section, cmds in sections.items():

                rows = []

                for name, cmd in cmds:
                    name = name.ljust(max_len)
                    help = cmd.get_short_help_str(limit)
                    rows.append((name, help))

                if rows:
                    with formatter.section(section):
                        formatter.write_dl(rows)


@click.group(cls=OrderedGroup, help="Dropbox client for Linux and macOS.")
@click.version_option(version=__version__, message=__version__)
def main():
    pass


# ======================================================================================
# Core commands
# ======================================================================================

config_option = click.option(
    "-c",
    "--config-name",
    default="maestral",
    type=ConfigName(existing=False),
    is_eager=True,
    expose_value=True,
    help="Run command with the given configuration.",
)

existing_config_option = click.option(
    "-c",
    "--config-name",
    default="maestral",
    type=ConfigName(),
    is_eager=True,
    expose_value=True,
    help="Run command with the given configuration.",
)


@main.command(section="Core Commands", help="Start the sync daemon.")
@click.option(
    "--foreground",
    "-f",
    is_flag=True,
    default=False,
    help="Start Maestral in the foreground.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Print log messages to stderr.",
)
@config_option
@convert_py_errors
def start(foreground: bool, verbose: bool, config_name: str) -> None:

    import threading
    from .daemon import (
        MaestralProxy,
        start_maestral_daemon,
        start_maestral_daemon_process,
        wait_for_startup,
        is_running,
        Start,
        CommunicationError,
    )

    check_for_updates()

    if is_running(config_name):
        click.echo("Daemon is already running.")
        return

    @convert_py_errors
    def startup_dialog():

        try:
            wait_for_startup(config_name)
        except CommunicationError:
            return

        m = MaestralProxy(config_name)

        if m.pending_link:
            link_dialog(m)

        if m.pending_dropbox_folder:
            path = select_dbx_path_dialog(config_name, allow_merge=True)

            while True:
                try:
                    m.create_dropbox_directory(path)
                    break
                except OSError:
                    cli.warn(
                        "Could not create folder. Please make sure that you have "
                        "permissions to write to the selected location or choose a "
                        "different location."
                    )

            include_all = cli.confirm("Would you like sync all folders?")

            if not include_all:
                # get all top-level Dropbox folders
                cli.info("Loading...")
                entries = m.list_folder("/", recursive=False)

                names = [
                    cast(str, e["name"])
                    for e in entries
                    if e["type"] == "FolderMetadata"
                ]

                choices = cli.select_multiple(
                    "Choose which folders to include", options=names
                )

                excluded_paths = [
                    f"/{name}"
                    for index, name in enumerate(names)
                    if index not in choices
                ]

                m.excluded_items = excluded_paths

            cli.ok("Setup completed. Starting sync.")

        m.start_sync()

    t = threading.Thread(target=startup_dialog)
    t.start()

    if foreground:
        start_maestral_daemon(config_name, log_to_stderr=verbose)
    else:
        cli.echo("Starting Maestral...", nl=False)

        res = start_maestral_daemon_process(config_name)

        if res == Start.Ok:
            cli.echo("\rStarting Maestral...        " + OK)
        elif res == Start.AlreadyRunning:
            cli.echo("\rStarting Maestral...        " + "Already running.")
        else:
            cli.echo("\rStarting Maestral...        " + FAILED)
            cli.echo("Please check logs for more information.")
            return


@main.command(section="Core Commands", help="Stop the sync daemon.")
@existing_config_option
def stop(config_name: str) -> None:
    stop_daemon_with_cli_feedback(config_name)


@main.command(section="Core Commands", help="Run the GUI if installed.")
@config_option
def gui(config_name: str) -> None:

    from packaging.version import Version
    from packaging.requirements import Requirement

    try:
        from importlib.metadata import entry_points, requires, version  # type: ignore
    except ImportError:
        from importlib_metadata import entry_points, requires, version  # type: ignore

    # find all "maestral_gui" entry points registered by other packages
    gui_entry_points = entry_points().get("maestral_gui")

    if not gui_entry_points or len(gui_entry_points) == 0:
        raise cli.CliException(
            "No maestral GUI installed. Please run 'pip3 install maestral[gui]'."
        )

    # check if 1st party defaults "maestral_cocoa" or "maestral_qt" are installed
    default_gui = "maestral_cocoa" if sys.platform == "darwin" else "maestral_qt"
    default_entry_point = next(
        (e for e in gui_entry_points if e.name == default_gui), None
    )

    if default_entry_point:
        # check gui requirements
        requirements = [Requirement(r) for r in requires("maestral")]  # type: ignore

        for r in requirements:
            if r.marker and r.marker.evaluate({"extra": "gui"}):
                version_str = version(r.name)
                if not r.specifier.contains(Version(version_str), prereleases=True):
                    raise cli.CliException(
                        f"{r.name}{r.specifier} required but you have {version_str}"
                    )

        # load entry point
        run = default_entry_point.load()

    else:
        # load any 3rd party GUI
        fallback_entry_point = next(iter(gui_entry_points))
        run = fallback_entry_point.load()

    run(config_name)


@main.command(section="Core Commands", help="Pause syncing.")
@existing_config_option
def pause(config_name: str) -> None:

    from .daemon import MaestralProxy, CommunicationError

    try:
        with MaestralProxy(config_name) as m:
            m.stop_sync()
        cli.ok("Syncing paused.")
    except CommunicationError:
        cli.echo("Maestral daemon is not running.")


@main.command(section="Core Commands", help="Resume syncing.")
@existing_config_option
def resume(config_name: str) -> None:

    from .daemon import MaestralProxy, CommunicationError

    try:
        with MaestralProxy(config_name) as m:
            if not check_for_fatal_errors(m):
                m.start_sync()
                cli.ok("Syncing resumed.")

    except CommunicationError:
        cli.echo("Maestral daemon is not running.")


@main.group(section="Core Commands", help="Link, unlink and view the Dropbox account.")
def auth():
    pass


@auth.command(name="link", help="Link a new Dropbox account.")
@click.option(
    "--relink",
    "-r",
    is_flag=True,
    default=False,
    help="Relink to the existing account. Keeps the sync state.",
)
@config_option
@convert_py_errors
def auth_link(relink: bool, config_name: str) -> None:

    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:

        if m.pending_link or relink:
            link_dialog(m)
        else:
            cli.echo(
                "Maestral is already linked. Use '-r' to relink to the same "
                "account or specify a new config name with '-c'."
            )


@auth.command(
    name="unlink",
    help="""
Unlink your Dropbox account.

If Maestral is running, it will be stopped before unlinking.
""",
)
@click.option(
    "--yes", "-Y", is_flag=True, default=False, help="Skip confirmation prompt."
)
@existing_config_option
@convert_py_errors
def auth_unlink(yes: bool, config_name: str) -> None:

    if not yes:
        yes = cli.confirm("Are you sure you want unlink your account?", default=False)

    if yes:
        from .main import Maestral

        stop_daemon_with_cli_feedback(config_name)
        m = Maestral(config_name)
        m.unlink()

        cli.ok("Unlinked Maestral.")


@auth.command(name="status", help="View authentication status.")
@existing_config_option
def auth_status(config_name: str) -> None:

    from .config import MaestralConfig, MaestralState

    conf = MaestralConfig(config_name)
    state = MaestralState(config_name)

    email = state.get("account", "email")
    account_type = state.get("account", "type").capitalize()
    dbid = conf.get("account", "account_id")

    cli.echo("")
    cli.echo(f"Email:         {email}")
    cli.echo(f"Account-type:  {account_type}")
    cli.echo(f"Dropbox-ID:    {dbid}")
    cli.echo("")


@main.group(section="Core Commands", help="Create and manage shared links.")
def sharelink():
    pass


@sharelink.command(name="create", help="Create a shared link for a file or folder.")
@click.argument("dropbox_path", type=DropboxPath())
@click.option(
    "-p",
    "--password",
    help="Optional password for the link.",
)
@click.option(
    "-e",
    "--expiry",
    metavar="DATE",
    type=click.DateTime(formats=["%Y-%m-%d", "%Y-%m-%dT%H:%M", "%Y-%m-%d %H:%M"]),
    help="Expiry time for the link (e.g. '2025-07-24 20:50').",
)
@existing_config_option
@convert_py_errors
def sharelink_create(
    dropbox_path: str,
    password: str,
    expiry: Optional["datetime"],
    config_name: str,
) -> None:

    from .daemon import MaestralProxy

    expiry_dt: Optional[float]

    if expiry:
        expiry_dt = expiry.timestamp()
    else:
        expiry_dt = None

    if password:
        visibility = "password"
    else:
        visibility = "public"

    with MaestralProxy(config_name, fallback=True) as m:
        link_info = m.create_shared_link(dropbox_path, visibility, password, expiry_dt)

    cli.echo(link_info["url"])


@sharelink.command(name="revoke", help="Revoke a shared link.")
@click.argument("url")
@existing_config_option
@convert_py_errors
def sharelink_revoke(url: str, config_name: str) -> None:

    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:
        m.revoke_shared_link(url)

    cli.confirm("Revoked shared link.")


@sharelink.command(
    name="list", help="List shared links for a path or all shared links."
)
@click.argument("dropbox_path", required=False, type=DropboxPath())
@existing_config_option
@convert_py_errors
def sharelink_list(dropbox_path: Optional[str], config_name: str) -> None:

    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:
        links = m.list_shared_links(dropbox_path)

    link_table = cli.Table(["URL", "Item", "Access", "Expires"])

    for link in links:
        url = cast(str, link["url"])
        file_name = cast(str, link["name"])
        visibility = cast(str, link["link_permissions"]["resolved_visibility"][".tag"])

        dt_field: cli.Field

        if "expires" in link:
            expires = cast(str, link["expires"])
            dt_field = cli.DateField(_datetime_from_iso_str(expires))
        else:
            dt_field = cli.TextField("-")

        link_table.append([url, file_name, visibility, dt_field])

    cli.echo("")
    link_table.echo()
    cli.echo("")


# ======================================================================================
# Information commands
# ======================================================================================


@main.command(section="Information", help="Show the status of the daemon.")
@existing_config_option
@convert_py_errors
def status(config_name: str) -> None:

    from .daemon import MaestralProxy, CommunicationError

    check_for_updates()

    try:
        with MaestralProxy(config_name) as m:

            email = m.get_state("account", "email")
            account_type = m.get_state("account", "type").capitalize()
            usage = m.get_state("account", "usage")
            status_info = m.status
            n_errors = len(m.sync_errors)
            color = "red" if n_errors > 0 else "green"
            n_errors_str = click.style(str(n_errors), fg=color)

            cli.echo("")
            cli.echo(f"Account:      {email} ({account_type})")
            cli.echo(f"Usage:        {usage}")
            cli.echo(f"Status:       {status_info}")
            cli.echo(f"Sync errors:  {n_errors_str}")
            cli.echo("")

            check_for_fatal_errors(m)

            sync_errors = m.sync_errors

            if len(sync_errors) > 0:

                path_column = cli.Column(title="Path")
                message_column = cli.Column(title="Error", wraps=True)

                for error in sync_errors:
                    path_column.append(error["dbx_path"])
                    message_column.append("{title}. {message}".format(**error))

                table = cli.Table([path_column, message_column])

                table.echo()
                cli.echo("")

    except CommunicationError:
        cli.echo("Maestral daemon is not running.")


@main.command(
    section="Information",
    help="""
Show the sync status of a local file or folder.

Returned value will be 'uploading', 'downloading', 'up to date', 'error', or 'unwatched'
(for files outside of the Dropbox directory). This will always be 'unwatched' if syncing
is paused. This command can be used to for instance to query information for a plugin to
a file-manager.
""",
)
@click.argument("local_path", type=click.Path(exists=True, resolve_path=True))
@existing_config_option
def filestatus(local_path: str, config_name: str) -> None:

    from .daemon import MaestralProxy, CommunicationError

    try:
        with MaestralProxy(config_name) as m:
            stat = m.get_file_status(local_path)
            cli.echo(stat)

    except CommunicationError:
        cli.echo("unwatched")


@main.command(section="Information", help="Live view of all items being synced.")
@existing_config_option
@convert_py_errors
def activity(config_name: str) -> None:

    import curses
    from .utils import natural_size
    from .daemon import MaestralProxy, CommunicationError

    try:
        with MaestralProxy(config_name) as m:

            if check_for_fatal_errors(m):
                return

            def curses_loop(screen) -> None:  # no type hints for screen provided yet

                curses.use_default_colors()  # don't change terminal background
                screen.nodelay(1)  # sets `screen.getch()` to non-blocking

                while True:

                    height, width = screen.getmaxyx()

                    # create header
                    lines = [
                        f"Status: {m.status}, Sync errors: {len(m.sync_errors)}",
                        "",
                    ]

                    # create table
                    filenames = []
                    states = []
                    col_len = 4

                    for event in m.get_activity(limit=height - 3):

                        dbx_path = cast(str, event["dbx_path"])
                        direction = cast(str, event["direction"])
                        state = cast(str, event["status"])
                        size = cast(int, event["size"])
                        completed = cast(int, event["completed"])

                        filename = os.path.basename(dbx_path)
                        filenames.append(filename)

                        arrow = "↓" if direction == "down" else "↑"

                        if completed > 0:
                            done_str = natural_size(completed, sep=False)
                            todo_str = natural_size(size, sep=False)
                            states.append(f"{done_str}/{todo_str} {arrow}")
                        else:
                            if state == "syncing" and direction == "up":
                                states.append("uploading")
                            elif state == "syncing" and direction == "down":
                                states.append("downloading")
                            else:
                                states.append(state)

                        col_len = max(len(filename), col_len)

                    for name, state in zip(filenames, states):  # create rows
                        lines.append(name.ljust(col_len + 2) + state)

                    # print to console screen
                    screen.clear()
                    try:
                        screen.addstr("\n".join(lines))
                    except curses.error:
                        pass
                    screen.refresh()

                    # abort when user presses 'q', refresh otherwise
                    key = screen.getch()
                    if key == ord("q"):
                        break
                    elif key < 0:
                        time.sleep(1)

            # enter curses event loop
            curses.wrapper(curses_loop)

    except CommunicationError:
        cli.echo("Maestral daemon is not running.")


@main.command(section="Information", help="Show recently changed or added files.")
@existing_config_option
@convert_py_errors
def history(config_name: str) -> None:

    from datetime import datetime
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:
        events = m.get_history()

    table = cli.Table(
        [
            cli.Column("Path", elide=cli.Elide.Leading),
            cli.Column("Change"),
            cli.Column("Time"),
        ]
    )

    for event in events:

        dbx_path = cast(str, event["dbx_path"])
        change_type = cast(str, event["change_type"])
        change_time_or_sync_time = cast(float, event["change_time_or_sync_time"])
        dt = datetime.fromtimestamp(change_time_or_sync_time)

        table.append([dbx_path, change_type, dt])

    cli.echo("")
    table.echo()
    cli.echo("")


@main.command(section="Information", help="List contents of a Dropbox directory.")
@click.argument("dropbox_path", type=DropboxPath(), default="")
@click.option(
    "-l",
    "--long",
    is_flag=True,
    default=False,
    help="Show output in long format with metadata.",
)
@click.option(
    "-d",
    "--include-deleted",
    is_flag=True,
    default=False,
    help="Include deleted items in listing.",
)
@existing_config_option
@convert_py_errors
def ls(long: bool, dropbox_path: str, include_deleted: bool, config_name: str) -> None:

    from .utils import natural_size
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:

        cli.echo("Loading...\r", nl=False)

        entries_iter = m.list_folder_iterator(
            dropbox_path,
            recursive=False,
            include_deleted=include_deleted,
        )

        if long:

            to_short_type = {
                "FileMetadata": "file",
                "FolderMetadata": "folder",
                "DeletedMetadata": "deleted",
            }

            table = cli.Table(
                columns=[
                    cli.Column("Name"),
                    cli.Column("Type"),
                    cli.Column("Size", align=cli.Align.Right),
                    cli.Column("Shared"),
                    cli.Column("Syncing"),
                    cli.Column("Last Modified"),
                ]
            )

            for entries in entries_iter:

                for entry in entries:

                    item_type = to_short_type[cast(str, entry["type"])]
                    name = cast(str, entry["name"])
                    path_lower = cast(str, entry["path_lower"])

                    text = "shared" if "sharing_info" in entry else "private"
                    color = "bright_black" if text == "private" else None
                    shared_field = cli.TextField(text, fg=color)

                    excluded_status = m.excluded_status(path_lower)
                    color = "green" if excluded_status == "included" else None
                    text = "✓" if excluded_status == "included" else excluded_status
                    excluded_field = cli.TextField(text, fg=color)

                    if "size" in entry:
                        size = natural_size(cast(float, entry["size"]))
                    else:
                        size = "-"

                    dt_field: cli.Field

                    if "client_modified" in entry:
                        cm = cast(str, entry["client_modified"])
                        dt_field = cli.DateField(_datetime_from_iso_str(cm))
                    else:
                        dt_field = cli.TextField("-")

                    table.append(
                        [name, item_type, size, shared_field, excluded_field, dt_field]
                    )

            cli.echo(" " * 15)
            table.echo()
            cli.echo(" " * 15)

        else:

            grid = cli.Grid()

            for entries in entries_iter:
                for entry in entries:
                    name = cast(str, entry["name"])
                    color = "blue" if entry["type"] == "DeletedMetadata" else None

                    grid.append(cli.TextField(name, fg=color))

            grid.echo()


@main.command(section="Information", help="List all configured Dropbox accounts.")
@click.option(
    "--clean",
    is_flag=True,
    default=False,
    help="Remove config files without a linked account.",
)
def config_files(clean: bool) -> None:

    from .daemon import is_running
    from .config import (
        MaestralConfig,
        MaestralState,
        list_configs,
        remove_configuration,
    )

    if clean:

        # Clean up stale config files.

        for name in list_configs():
            conf = MaestralConfig(name)
            path = conf.get_config_fpath()
            dbid = conf.get("account", "account_id")

            if dbid == "" and not is_running(name):
                remove_configuration(name)
                cli.echo(f"Removed: {path}")

    else:
        # Display config files.
        names = list_configs()
        emails = []
        paths = []

        for name in names:
            config = MaestralConfig(name)
            state = MaestralState(name)

            emails.append(state.get("account", "email"))
            paths.append(config.get_config_fpath())

        table = cli.Table(
            [
                cli.Column("Config name", names),
                cli.Column("Account", emails),
                cli.Column("Path", paths, elide=cli.Elide.Leading),
            ]
        )

        cli.echo("")
        table.echo()
        cli.echo("")


# ======================================================================================
# Settings
# ======================================================================================


@main.command(
    section="Settings",
    help="""
Automatically start the sync daemon on login.

A systemd or launchd service will be created to start a sync daemon for the given
configuration on user login.
""",
)
@click.option("--yes", "-Y", is_flag=True, default=False)
@click.option("--no", "-N", is_flag=True, default=False)
@existing_config_option
def autostart(yes: bool, no: bool, config_name: str) -> None:

    from .autostart import AutoStart

    auto_start = AutoStart(config_name)

    if not auto_start.implementation:
        cli.echo(
            "Autostart is currently not supported for your platform.\n"
            "Autostart requires systemd on Linux or launchd on macOS."
        )
        return

    if yes or no:
        if yes:
            auto_start.enable()
            cli.ok("Enabled start on login.")
        else:
            auto_start.disable()
            cli.ok("Disabled start on login.")
    else:
        if auto_start.enabled:
            cli.echo("Autostart is enabled. Use -N to disable.")
        else:
            cli.echo("Autostart is disabled. Use -Y to enable.")


@main.group(section="Settings", help="View and manage excluded folders.")
def excluded():
    pass


@excluded.command(name="list", help="List all excluded files and folders.")
@existing_config_option
def excluded_list(config_name: str) -> None:

    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:

        excluded_items = m.excluded_items
        excluded_items.sort()

        if len(excluded_items) == 0:
            cli.echo("No excluded files or folders.")
        else:
            for item in excluded_items:
                cli.echo(item)


@excluded.command(
    name="add",
    help="Add a file or folder to the excluded list and re-sync.",
)
@click.argument("dropbox_path", type=DropboxPath())
@existing_config_option
@convert_py_errors
def excluded_add(dropbox_path: str, config_name: str) -> None:

    from .daemon import MaestralProxy

    if dropbox_path == "/":
        raise cli.CliException("Cannot exclude the root directory.")

    with MaestralProxy(config_name, fallback=True) as m:
        m.exclude_item(dropbox_path)
        cli.ok(f"Excluded '{dropbox_path}'.")


@excluded.command(
    name="remove",
    help="""
Remove a file or folder from the excluded list and re-sync.

It is safe to call this method with items which have already been included, they will
not be downloaded again. If the given path lies inside an excluded folder, the parent
folder will be included as well (but no other items inside it).
""",
)
@click.argument("dropbox_path", type=DropboxPath())
@existing_config_option
@convert_py_errors
def excluded_remove(dropbox_path: str, config_name: str) -> None:

    from .daemon import MaestralProxy, CommunicationError

    if dropbox_path == "/":
        return cli.echo("The root directory is always included")

    try:
        with MaestralProxy(config_name) as m:
            m.include_item(dropbox_path)
            cli.ok(f"Included '{dropbox_path}'. Now downloading...")

    except CommunicationError:
        raise cli.CliException("Daemon must be running to download folders.")


@main.group(section="Settings", help="Manage desktop notifications.")
def notify():
    pass


@notify.command(
    name="level",
    help="Get or set the level for desktop notifications.",
)
@click.argument(
    "level_name",
    required=False,
    type=click.Choice(["ERROR", "SYNCISSUE", "FILECHANGE"], case_sensitive=False),
)
@existing_config_option
def notify_level(level_name: str, config_name: str) -> None:

    from . import notify as _notify
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:
        if level_name:
            m.notification_level = _notify.level_name_to_number(level_name)
            cli.ok(f"Notification level set to {level_name}.")
        else:
            level_name = _notify.level_number_to_name(m.notification_level)
            cli.echo(f"Notification level: {level_name}.")


@notify.command(
    name="snooze",
    help="Snooze desktop notifications of file changes.",
)
@click.argument("minutes", type=click.IntRange(min=0))
@existing_config_option
def notify_snooze(minutes: int, config_name: str) -> None:

    from .daemon import MaestralProxy, CommunicationError

    try:
        with MaestralProxy(config_name) as m:
            m.notification_snooze = minutes
    except CommunicationError:
        cli.echo("Maestral daemon is not running.")
    else:
        if minutes > 0:
            cli.ok(
                f"Notifications snoozed for {minutes} min. Set snooze to 0 to reset."
            )
        else:
            cli.ok("Notifications enabled.")


# ======================================================================================
# Maintenance
# ======================================================================================


@main.command(section="Maintenance", help="Move the local Dropbox folder.")
@click.argument("new_path", required=False, type=click.Path(writable=True))
@existing_config_option
def move_dir(new_path: str, config_name: str) -> None:

    from .daemon import MaestralProxy

    new_path = new_path or select_dbx_path_dialog(config_name)

    with MaestralProxy(config_name, fallback=True) as m:
        m.move_dropbox_directory(new_path)

    cli.ok(f"Dropbox folder moved to {new_path}.")


@main.command(
    section="Maintenance",
    help="""
Rebuild the sync index.

Rebuilding may take several minutes, depending on the size of your Dropbox.
""",
)
@click.option(
    "--yes", "-Y", is_flag=True, default=False, help="Skip confirmation prompt."
)
@existing_config_option
@convert_py_errors
def rebuild_index(yes: bool, config_name: str) -> None:

    import textwrap
    import shutil
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:

        width, height = shutil.get_terminal_size()

        msg = textwrap.fill(
            "Rebuilding the index may take several minutes, depending on the size of "
            "your Dropbox. Any changes to local files will be synced once rebuilding "
            "has completed. If you stop the daemon during the process, rebuilding will "
            "start again on the next launch.\nIf the daemon is not currently running, "
            "a rebuild will be scheduled for the next startup.",
            width=width,
        )

        cli.echo(msg + "\n")

        if yes or cli.confirm("Do you want to continue?", default=False):

            m.rebuild_index()

            if m._is_fallback:
                cli.ok("Daemon is not running. Rebuilding scheduled for next startup.")
            else:
                cli.ok("Rebuilding now. Run 'maestral status' to view progress.")


@main.command(section="Maintenance", help="List old file revisions.")
@click.argument("dropbox_path", type=DropboxPath())
@click.option(
    "-l",
    "--limit",
    help="Maximum number of revs to list.",
    show_default=True,
    type=click.IntRange(min=1, max=100),
    default=10,
)
@existing_config_option
@convert_py_errors
def revs(dropbox_path: str, limit: int, config_name: str) -> None:

    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:
        entries = m.list_revisions(dropbox_path, limit=limit)

    table = cli.Table(["Revision", "Modified Time"])

    for entry in entries:

        rev = cast(str, entry["rev"])
        dt = _datetime_from_iso_str(cast(str, entry["client_modified"]))

        table.append([cli.TextField(rev), cli.DateField(dt)])

    cli.echo("")
    table.echo()
    cli.echo("")


@main.command(
    section="Maintenance",
    help="""
Compare two revisions of a file.

If no revs are passed to the command, you can select the revisions interactively. If
only one rev is passed, it is compared to the local version of the file. The diff is
shown via a pager if longer 30 lines.

Warning: The specified revisions will be downloaded to temp files and loaded into memory
to generate the diff. Depending on the file size, this may use significant disk space
and memory.
""",
)
@click.argument("dropbox_path", type=DropboxPath())
@click.option(
    "-v",
    "--rev",
    help="Revisions to compare (multiple allowed).",
    multiple=True,
    default=[],
)
@click.option("--no-color", help="Don't use colors for the diff.", is_flag=True)
@click.option("--no-pager", help="Don't use a pager for output.", is_flag=True)
@click.option(
    "-l",
    "--limit",
    help="Maximum number of revs to list.",
    show_default=True,
    type=click.IntRange(min=1, max=100),
    default=10,
)
@convert_py_errors
@existing_config_option
def diff(
    dropbox_path: str,
    rev: List[str],
    no_color: bool,
    no_pager: bool,
    limit: int,
    config_name: str,
) -> None:

    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:

        # Ask for user input if revs are not provided as CLI arguments.
        if len(rev) == 0:
            entries = m.list_revisions(dropbox_path, limit=limit)

            for entry in entries:
                cm = cast(str, entry["client_modified"])
                field = cli.DateField(_datetime_from_iso_str(cm))
                entry["desc"] = field.format(40)[0]

            dbx_path = cast(str, entries[0]["path_display"])
            local_path = m.to_local_path(dbx_path)

            if osp.isfile(local_path):
                # prepend local version as an option
                entries.insert(0, {"desc": "local version", "rev": None})

            index_base = cli.select(
                message="New revision:",
                options=[e["desc"] for e in entries],
                hint="(↓ to see more)" if len(entries) > 6 else "",
            )

            if index_base == len(entries) - 1:
                cli.warn(
                    "Oldest revision selected, unable to find anything to compare."
                )
                return

            comparable_versions = entries[index_base + 1 :]
            index_new = cli.select(
                message="Old revision:",
                options=[e["desc"] for e in comparable_versions],
                hint="(↓ to see more)" if len(comparable_versions) > 6 else "",
            )

            old_rev = entries[index_new + index_base + 1]["rev"]
            new_rev = entries[index_base]["rev"]
        elif len(rev) == 1:
            old_rev = rev[0]
            new_rev = None
        elif len(rev) == 2:
            old_rev = rev[0]
            new_rev = rev[1]
        elif len(rev) > 2:
            cli.warn("You can only compare two revisions at a time.")
            return

        # Download up to two revisions to a local temporary folder
        # and compare them with a 'diff'. Only text files are supported.
        # If an unknown file type was found, everything that doesn't match
        # 'text/*', an error message gets printed.

        click.echo("Loading ...\r", nl=False)

        diff_output = m.get_file_diff(old_rev, new_rev)

        if len(diff_output) == 0:
            click.echo("There are no changes between the two revisions.")
            return

        def color(ind: int, line: str) -> str:
            """
            Color diff lines.
            Inspiration for colors was taken from the
            well known command 'git diff'.
            """

            if ind < 2:
                line = click.style(line, bold=True)
            elif line.startswith("+"):
                line = click.style(line, fg="green")
            elif line.startswith("-"):
                line = click.style(line, fg="red")
            # Don't highlight these in the intro.
            elif line.startswith("@@ "):
                line = click.style(line, fg="cyan")
            return line

        # Color the lines.
        if not no_color:
            diff_output = [color(i, l) for i, l in enumerate(diff_output)]

        # Enter pager if diff is too long
        if len(diff_output) > 30 and not no_pager:
            click.echo_via_pager("".join(diff_output))
        else:
            click.echo("".join(diff_output))


@main.command(
    section="Maintenance",
    help="""
Restore a previous version of a file.

If no revision number is given, old revisions will be listed.
""",
)
@click.argument("dropbox_path", type=DropboxPath())
@click.option("-v", "--rev", help="Revision to restore.", default="")
@click.option(
    "-l",
    "--limit",
    help="Maximum number of revs to list.",
    show_default=True,
    type=click.IntRange(min=1, max=100),
    default=10,
)
@existing_config_option
@convert_py_errors
def restore(dropbox_path: str, rev: str, limit: int, config_name: str) -> None:

    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:

        if not rev:
            cli.echo("Loading...\r", nl=False)
            entries = m.list_revisions(dropbox_path, limit=limit)
            dates = []
            for entry in entries:
                cm = cast(str, entry["client_modified"])
                field = cli.DateField(_datetime_from_iso_str(cm))
                dates.append(field.format(40)[0])

            index = cli.select(
                message="Select a version to restore:",
                options=dates,
                hint="(↓ to see more)" if len(entries) > 6 else "",
            )
            rev = cast(str, entries[index]["rev"])

        m.restore(dropbox_path, rev)

    cli.ok(f'Restored {rev} to "{dropbox_path}"')


@main.group(section="Maintenance", help="View and manage the log.")
def log():
    pass


@log.command(name="show", help="Print logs to the console.")
@click.option(
    "--external", "-e", is_flag=True, default=False, help="Open in external program."
)
@existing_config_option
def log_show(external: bool, config_name: str) -> None:

    from .utils.appdirs import get_log_path

    log_file = get_log_path("maestral", config_name + ".log")

    if external:
        res = click.launch(log_file)
    else:
        try:
            with open(log_file) as f:
                text = f.read()
            click.echo_via_pager(text)
        except OSError:
            res = 1
        else:
            res = 0

    if res > 0:
        raise cli.CliException(f"Could not open log file at '{log_file}'")


@log.command(name="clear", help="Clear the log files.")
@existing_config_option
def log_clear(config_name: str) -> None:

    from .utils.appdirs import get_log_path

    log_dir = get_log_path("maestral")
    log_name = config_name + ".log"

    log_files = []

    for file_name in os.listdir(log_dir):
        if file_name.startswith(log_name):
            log_files.append(os.path.join(log_dir, file_name))

    try:
        for file in log_files:
            open(file, "w").close()
        cli.ok("Cleared log files.")
    except FileNotFoundError:
        cli.ok("Cleared log files.")
    except OSError:
        raise cli.CliException(
            f"Could not clear log at '{log_dir}'. " f"Please try to delete it manually"
        )


@log.command(name="level", help="Get or set the log level.")
@click.argument(
    "level_name",
    required=False,
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
)
@existing_config_option
def log_level(level_name: str, config_name: str) -> None:

    import logging
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:
        if level_name:
            m.log_level = cast(int, getattr(logging, level_name))
            cli.ok(f"Log level set to {level_name}.")
        else:
            level_name = logging.getLevelName(m.log_level)
            cli.echo(f"Log level: {level_name}")


@main.group(
    section="Maintenance",
    help="""
Direct access to config values.

Warning: Changing some config values must be accompanied by maintenance tasks. For
example, changing the config value for the Dropbox location needs to be accompanied by
actually moving the folder. This command only gets / sets the value in the config file.
Most changes will also require a restart of the daemon to become effective.

Use the commands from the Settings section instead wherever possible. They will take
effect immediately, perform accompanying tasks for you, and never leave the daemon in an
inconsistent state.

Currently available config keys are:

\b
- path: the location of the local Dropbox folder
- excluded_items: list of files or folders excluded by selective sync
- account_id: the ID of the linked Dropbox account
- notification_level: the level for desktop notifications
- log_level: the log level.
- update_notification_interval: interval in secs to check for updates
- keyring: the keyring backend to use (full path of the class)
- reindex_interval: the interval in seconds for full reindexing
- max_cpu_percent: maximum CPU usage target per core
- keep_history: the sync history to keep in seconds
- upload: if upload sync is enabled
- download: if download sync is enabled
""",
)
def config():
    pass


@config.command(name="get", help="Print the value of a given configuration key.")
@click.argument("key", type=ConfigKey())
@config_option
def config_get(key: str, config_name: str) -> None:

    from .config import MaestralConfig
    from .config.main import KEY_SECTION_MAP
    from .daemon import MaestralProxy, CommunicationError

    # Check if the config key exists in any section.
    section = KEY_SECTION_MAP.get(key, "")

    if not section:
        raise cli.CliException(f"'{key}' is not a valid configuration key.")

    try:
        with MaestralProxy(config_name) as m:
            value = m.get_conf(section, key)
    except CommunicationError:
        value = MaestralConfig(config_name).get(section, key)

    cli.echo(value)


@config.command(
    name="set",
    help="""
Update configuration with a value for the given key.

Values will be cast to the proper type, raising an error where this is not possibly. For
instance, setting a boolean config value to 1 will actually set it to True.
""",
)
@click.argument("key", type=ConfigKey())
@click.argument("value")
@config_option
@convert_py_errors
def config_set(key: str, value: str, config_name: str) -> None:

    import ast
    from .config import MaestralConfig
    from .config.main import KEY_SECTION_MAP, DEFAULTS_CONFIG
    from .daemon import MaestralProxy, CommunicationError

    section = KEY_SECTION_MAP.get(key, "")

    if not section:
        raise cli.CliException(f"'{key}' is not a valid configuration key.")

    default_value = DEFAULTS_CONFIG[section][key]

    if isinstance(default_value, str):
        py_value = value
    else:
        try:
            py_value = ast.literal_eval(value)
        except (SyntaxError, ValueError):
            py_value = value

    try:
        with MaestralProxy(config_name) as m:
            m.set_conf(section, key, py_value)
    except CommunicationError:
        MaestralConfig(config_name).set(section, key, py_value)


@config.command(name="show", help="Show all config keys and values")
@click.option("--no-pager", help="Don't use a pager for output.", is_flag=True)
@config_option
def config_show(no_pager: bool, config_name: str) -> None:

    import io
    from .config import MaestralConfig

    conf = MaestralConfig(config_name)

    with io.StringIO() as fp:
        conf.write(fp)
        if no_pager:
            click.echo(fp.getvalue())
        else:
            click.echo_via_pager(fp.getvalue())


@main.command(
    section="Maintenance",
    help="""
Generate completion script for your shell.

This command can generate shell completion scripts for bash, zsh or fish. Follow the
instructions below for your shell to load the resulting script. The exact config file
locations might vary based on your system. Make sure to restart your
shell before testing whether completions are working.

### bash

You can enable shell completion for all users by generating and saving the script as
follows:

\b
    maestral completion bash > /usr/share/bash-completion/completions/maestral

To enable shell completion for the current user only, save the script in a location of
your choice, for example `~/.local/completions/maestral`, and source it in `~/.bashrc`
by adding the line:

\b
    . ~/.local/completions/maestral

### zsh

Generate a `_maestral` completion script and put it somewhere in your `$fpath`. For
example:

\b
    maestral completion zsh > /usr/local/share/zsh/site-functions/_maestral

You can also save the completion script in a location of your choice and source it
in `~/.zshrc`. Ensure that the following is present in your `~/.zshrc`:

\b
    autoload -Uz compinit && compinit

### fish

Generate and save a `maestral.fish` completion script as follows. For all users:

\b
    maestral completion fish > /usr/share/fish/vendor_completions.d/maestral.fish

For the current user only:

\b
    maestral completion fish > ~/.config/fish/completions/maestral.fish

""",
)
@click.argument("shell", type=click.Choice(["bash", "zsh", "fish"]))
def completion(shell: str) -> None:

    from click.shell_completion import get_completion_class

    comp_cls = get_completion_class(shell)

    if comp_cls is None:
        cli.warn(f"{shell} shell is currently not supported")
        return

    comp = comp_cls(main, {}, "maestral", "_MAESTRAL_COMPLETE")

    try:
        click.echo(comp.source())
    except RuntimeError as exc:
        cli.warn(exc.args[0])
